"""
Language file verification tests.

These are spot-check sanity checks only.  Ideally the tests would
verify that the language definition files are in fact **complete**,
i.e. contain all the necessary fields, but for now this will suffice.
Lute will double-check that the files are valid anyway.
"""

import unittest
import re
from glob import glob
import os
import yaml


class TestVerifyFiles(unittest.TestCase):
    "Verify files."

    def test_definition_files_are_valid_yaml(self):
        "Files must be valid yaml."
        thisdir = os.path.dirname(__file__)
        langglob = os.path.join(thisdir, "**", "definition.yaml")
        for filename in glob(langglob):
            print(filename)
            with open(filename, "r", encoding="utf-8") as f:
                try:
                    d = yaml.safe_load(f)
                except Exception as e:  # pylint: disable=broad-exception-caught
                    self.fail(f"ERROR: {filename} invalid: {e}")
                self.assertTrue(d["name"] is not None, f"{filename} missing name")

    def test_story_files_contain_title(self):
        "Must have # title: at the top."
        thisdir = os.path.dirname(__file__)
        storyglob = os.path.join(thisdir, "**", "*.txt")
        for filename in glob(storyglob):
            print(filename)
            with open(filename, "r", encoding="utf-8") as f:
                content = f.read()
            retitle = r"title:\s*(.*)\n"
            t = re.search(retitle, content)
            msg = f"{filename} missing '# title:' comment, {retitle}"
            self.assertTrue(t is not None, msg)
            title = t.group(1).strip()
            self.assertTrue(title != "", f"{filename} missing title")


if __name__ == "__main__":
    unittest.main()



"""
App configuration.
"""

import os
import yaml
from platformdirs import PlatformDirs


class AppConfig:  # pylint: disable=too-many-instance-attributes
    """
    Configuration wrapper around yaml file.

    Adds various properties for lint-time checking.
    """

    def __init__(self, config_file_path):
        """
        Load the required configuration file.
        """
        self._load_config(config_file_path)

    def _load_config(self, config_file_path):
        """
        Load and validate the config file.
        """
        with open(config_file_path, "r", encoding="utf-8") as cf:
            config = yaml.safe_load(cf)

        if not isinstance(config, dict):
            raise RuntimeError(
                f"File at {config_file_path} is invalid or is not a yaml dictionary."
            )

        self.env = config.get("ENV", None)
        if self.env not in ["prod", "dev"]:
            raise ValueError(f"ENV must be prod or dev, was {self.env}.")

        self.is_docker = bool(config.get("IS_DOCKER", False))

        # Database name.
        self.dbname = config.get("DBNAME", None)
        if self.dbname is None:
            raise ValueError("Config file must have 'DBNAME'")

        # Various invoke tasks in /tasks.py check if the database is a
        # test_ db prior to running some destructive action.
        self.is_test_db = self.dbname.startswith("test_")

        # Path to user data.
        self.datapath = config.get("DATAPATH", self._get_appdata_dir())
        self.plugin_datapath = os.path.join(self.datapath, "plugins")
        self.userimagespath = os.path.join(self.datapath, "userimages")
        self.useraudiopath = os.path.join(self.datapath, "useraudio")
        self.userthemespath = os.path.join(self.datapath, "userthemes")
        self.temppath = os.path.join(self.datapath, "temp")
        self.dbfilename = os.path.join(self.datapath, self.dbname)

        # Path to db backup.
        # When Lute starts up, it backs up the db
        # if migrations are going to be applied, just in case.
        # Hidden directory as a hint to the the user that
        # this is a system dir.
        self.system_backup_path = os.path.join(self.datapath, ".system_db_backups")

        # Default backup path for user, can be overridden in settings.
        self.default_user_backup_path = config.get(
            "BACKUP_PATH", os.path.join(self.datapath, "backups")
        )

    def _get_appdata_dir(self):
        "Get user's appdata directory from platformdirs."
        dirs = PlatformDirs("Lute3", "Lute3")
        return dirs.user_data_dir

    @property
    def sqliteconnstring(self):
        "Full sqlite connection string."
        return f"sqlite:///{self.dbfilename}"

    @staticmethod
    def configdir():
        "Return the path to the configuration file directory."
        return os.path.dirname(os.path.realpath(__file__))

    @staticmethod
    def default_config_filename():
        "Return the path to the default configuration file."
        thisdir = AppConfig.configdir()
        return os.path.join(thisdir, "config.yml")



"""
Reading helpers.
"""

from collections import defaultdict
from datetime import datetime
import functools
from lute.models.term import Term, Status
from lute.models.book import Text, WordsRead
from lute.models.repositories import BookRepository, UserSettingRepository
from lute.book.stats import Service as StatsService
from lute.read.render.service import Service as RenderService
from lute.read.render.calculate_textitems import get_string_indexes
from lute.term.model import Repository

# from lute.utils.debug_helpers import DebugTimer


class TermPopup:
    "Popup data for a term."

    # pylint: disable=too-many-instance-attributes
    def __init__(self, term):
        self.term = term
        self.term_text = self._clean(term.text)
        self.parents_text = ", ".join([self._clean(p.text) for p in term.parents])
        self.translation = self._clean(term.translation)
        self.romanization = self._clean(term.romanization)
        self.tags = [tt.text for tt in term.term_tags]
        self.flash = self._clean(term.get_flash_message())
        self.image = term.get_current_image()
        self.popup_image_data = self._get_popup_image_data()

        # Final data to include in popup.
        self.parents = []
        self.components = []

    def _clean(self, t):
        "Clean text for popup usage."
        zws = "\u200B"
        ret = (t or "").strip()
        ret = ret.replace(zws, "")
        ret = ret.replace("\n", "<br />")
        return ret

    @property
    def show(self):
        "Calc if should show.  Must be deferred as values can be changed."
        checks = [self.romanization != "", self.translation != "", len(self.tags) > 0]
        return len([b for b in checks if b]) > 0

    def term_and_parents_text(self):
        "Return term text with parents if any."
        ret = self.term_text
        if self.parents_text != "":
            ret = f"{ret} ({self.parents_text})"
        return ret

    def _get_popup_image_data(self):
        "Get images"
        # Don't include component images in the hover for now,
        # it can get confusing!
        # ref https://github.com/LuteOrg/lute-v3/issues/355
        terms = [self.term, *self.term.parents]
        images = [
            (t.get_current_image(), t.text) for t in terms if t.get_current_image()
        ]
        imageresult = defaultdict(list)
        for key, value in images:
            imageresult[key].append(self._clean(value))
        # Convert lists to comma-separated strings
        return {k: ", ".join(v) for k, v in imageresult.items()}


class Service:
    "Service."

    def __init__(self, session):
        self.session = session

    def mark_page_read(self, bookid, pagenum, mark_rest_as_known):
        "Mark page as read, record stats, rest as known."
        br = BookRepository(self.session)
        book = br.find(bookid)
        text = book.text_at_page(pagenum)
        d = datetime.now()
        text.read_date = d
        w = WordsRead(text, d, text.word_count)
        self.session.add(text)
        self.session.add(w)
        self.session.commit()
        if mark_rest_as_known:
            self.set_unknowns_to_known(text)

    def set_unknowns_to_known(self, text: Text):
        """
        Given a text, create new Terms with status Well-Known
        for any new Terms.
        """
        rs = RenderService(self.session)
        paragraphs = rs.get_paragraphs(text.text, text.book.language)
        self._save_new_status_0_terms(paragraphs)

        unknowns = [
            ti.term
            for para in paragraphs
            for sentence in para
            for ti in sentence
            if ti.is_word and ti.term.status == 0
        ]

        batch_size = 100
        i = 0

        for t in unknowns:
            t.status = Status.WELLKNOWN
            self.session.add(t)
            i += 1
            if i % batch_size == 0:
                self.session.commit()

        # Commit any remaining.
        self.session.commit()

    def bulk_status_update(self, text: Text, terms_text_array, new_status):
        """
        Given a text and list of terms, update or create new terms
        and set the status.
        """
        language = text.book.language
        repo = Repository(self.session)
        for term_text in terms_text_array:
            t = repo.find_or_new(language.id, term_text)
            t.status = new_status
            repo.add(t)
        repo.commit()

    def _save_new_status_0_terms(self, paragraphs):
        "Add status 0 terms for new textitems in paragraph."
        tis_with_new_terms = [
            ti
            for para in paragraphs
            for sentence in para
            for ti in sentence
            if ti.is_word and ti.term.id is None and ti.term.status == 0
        ]

        for ti in tis_with_new_terms:
            self.session.add(ti.term)
        self.session.commit()

    def start_reading(self, dbbook, pagenum):
        "Start reading a page in the book, getting paragraphs."

        text = dbbook.text_at_page(pagenum)
        text.load_sentences()

        svc = StatsService(self.session)
        svc.mark_stale(dbbook)
        dbbook.current_tx_id = text.id
        self.session.add(dbbook)
        self.session.add(text)
        self.session.commit()

        lang = text.book.language
        rs = RenderService(self.session)
        paragraphs = rs.get_paragraphs(text.text, lang)
        self._save_new_status_0_terms(paragraphs)

        return paragraphs

    def _sort_components(self, term, components):
        "Sort components by min position in string and length."
        component_and_pos = []
        for c in components:
            c_indices = [
                loc[1] for loc in get_string_indexes([c.text_lc], term.text_lc)
            ]

            # Sometimes the components aren't found
            # in the string, which makes no sense ...
            # ref https://github.com/LuteOrg/lute-v3/issues/474
            if len(c_indices) > 0:
                component_and_pos.append([c, min(c_indices)])

        def compare(a, b):
            # Lowest position (closest to front of string) sorts first.
            if a[1] != b[1]:
                return -1 if (a[1] < b[1]) else 1
            # Longest sorts first.
            alen = len(a[0].text)
            blen = len(b[0].text)
            return -1 if (alen > blen) else 1

        component_and_pos.sort(key=functools.cmp_to_key(compare))
        return [c[0] for c in component_and_pos]

    def get_popup_data(self, termid):
        "Get popup data, or None if popup shouldn't be shown."
        term = self.session.get(Term, termid)
        if term is None:
            return None

        repo = UserSettingRepository(self.session)
        show_components = int(repo.get_value("term_popup_show_components")) == 1
        components = []
        if show_components:
            rs = RenderService(self.session)
            components = [
                c
                for c in rs.find_all_Terms_in_string(term.text, term.language)
                if c.id != term.id and c.status != Status.UNKNOWN
            ]

        t = TermPopup(term)
        if (
            t.show is False
            and t.image is None
            and len(term.parents) == 0
            and len(components) == 0
        ):
            # Nothing to show."
            return None

        parent_data = [TermPopup(p) for p in term.parents]

        promote_parent_trans = int(
            repo.get_value("term_popup_promote_parent_translation")
        )
        if (promote_parent_trans == 1) and len(term.parents) == 1:
            ptrans = parent_data[0].translation
            if t.translation == "":
                t.translation = ptrans
            if t.translation == ptrans:
                parent_data[0].translation = ""

        component_data = [TermPopup(c) for c in self._sort_components(term, components)]

        t.parents = [p for p in parent_data if p.show]
        t.components = [c for c in component_data if c.show]
        return t



"""
/term service for routes to use
"""

from dataclasses import dataclass, field
from typing import List, Optional
from lute.models.term import Status
from lute.models.repositories import TermRepository, TermTagRepository
from lute.term.model import Repository


class TermServiceException(Exception):
    """
    Raised if something bad:

    - missing parent
    etc.
    """


# pylint: disable=too-many-instance-attributes
@dataclass
class BulkTermUpdateData:
    "Bulk updates"
    term_ids: List[int] = field(default_factory=list)
    lowercase_terms: bool = False
    remove_parents: bool = False
    parent_id: Optional[int] = None
    parent_text: Optional[str] = None
    change_status: bool = False
    status_value: Optional[int] = None
    add_tags: List[str] = field(default_factory=list)
    remove_tags: List[str] = field(default_factory=list)


class Service:
    "Service."

    def __init__(self, session):
        self.session = session

    def apply_bulk_updates(self, bulk_update_data):
        "Apply all updates."
        if len(bulk_update_data.term_ids) == 0:
            return

        parent = None
        repo = TermRepository(self.session)
        terms = [repo.find(tid) for tid in bulk_update_data.term_ids]

        lang_ids = list({term.language.id for term in terms})
        if len(lang_ids) > 1:
            raise TermServiceException("Terms not all the same language")

        # parent is found either by the ID, or if that returns None, by a text search.
        if bulk_update_data.parent_id is not None:
            parent = repo.find(bulk_update_data.parent_id)
        if parent is None and bulk_update_data.parent_text is not None:
            modelrepo = Repository(self.session)
            pmodel = modelrepo.find_or_new(lang_ids[0], bulk_update_data.parent_text)
            modelrepo.add(pmodel)
            modelrepo.commit()
            # Re-load it to get its id.  ... wasteful, not concerned at the moment.
            pmodel = modelrepo.find(lang_ids[0], bulk_update_data.parent_text)
            parent = repo.find(pmodel.id)

        ttrepo = TermTagRepository(self.session)
        add_tags = [ttrepo.find_or_create_by_text(a) for a in bulk_update_data.add_tags]
        remove_tags = [
            ttrepo.find_or_create_by_text(a) for a in bulk_update_data.remove_tags
        ]

        for term in terms:
            if bulk_update_data.lowercase_terms:
                term.text = term.text_lc
            if bulk_update_data.remove_parents:
                term.remove_all_parents()
                term.sync_status = False
            if parent is not None:
                term.remove_all_parents()
                term.add_parent(parent)
            if parent is not None and parent.status != Status.UNKNOWN:
                term.sync_status = True
                term.status = parent.status

            if (
                bulk_update_data.change_status is True
                and bulk_update_data.status_value is not None
            ):
                term.status = bulk_update_data.status_value

            for tag in add_tags:
                term.add_term_tag(tag)
            for tag in remove_tags:
                term.remove_term_tag(tag)

            self.session.add(term)
            self.session.commit()



"""
Testing management functions.

"management" is for global actions like clearing out the db.
"""

import pytest
from sqlalchemy import text
from lute.db import db
from lute.models.setting import UserSetting
from lute.models.repositories import UserSettingRepository
from lute.db.management import delete_all_data, add_default_user_settings
from tests.dbasserts import assert_record_count_equals


def test_wiping_db_clears_out_all_tables(app_context):
    """
    DB is wiped clean if requested ... settings are left!
    """
    old_user_settings = db.session.query(UserSetting).all()

    delete_all_data(db.session)
    tables = [
        "books",
        "bookstats",
        "booktags",
        "languages",
        "sentences",
        "tags",
        "tags2",
        "texts",
        "wordflashmessages",
        "wordimages",
        "wordparents",
        "words",
        "wordsread",
        "wordtags",
    ]
    for t in tables:
        assert_record_count_equals(t, 0, t)

    sql = "select * from settings where stkeytype='user'"
    assert_record_count_equals(sql, len(old_user_settings), "user settings remain")
    sql = "select * from settings where StKeyType = 'system'"
    assert_record_count_equals(sql, 0, "no system settings")


def test_can_get_backup_settings_when_db_is_wiped(app_context):
    "The backupsettings struct assumes certain things about the data."
    delete_all_data(db.session)
    repo = UserSettingRepository(db.session)
    bs = repo.get_backup_settings()
    assert bs.backup_enabled, "backup is back to being enabled"
    assert bs.backup_dir is not None, "default restored"


@pytest.fixture(name="us_repo")
def fixture_usersetting_repo(app_context):
    "Repo"
    r = UserSettingRepository(db.session)
    return r


def test_user_settings_loaded_with_defaults(us_repo):
    "Called on load."
    db.session.execute(text("delete from settings"))
    db.session.commit()
    assert us_repo.key_exists("backup_dir") is False, "key removed"
    add_default_user_settings(db.session, "blah")
    assert us_repo.key_exists("backup_dir") is True, "key created"

    # Check defaults
    b = us_repo.get_backup_settings()
    assert b.backup_enabled is True
    assert b.backup_dir is not None
    assert b.backup_auto is True
    assert b.backup_warn is True
    assert b.backup_count == 5


def test_user_settings_load_leaves_existing_values(us_repo):
    "Called on load."
    us_repo.set_value("backup_count", 17)
    db.session.commit()
    assert us_repo.get_value("backup_count") == "17"
    add_default_user_settings(db.session, "blah")
    b = us_repo.get_backup_settings()
    assert b.backup_count == 17, "still 17"



"""
Database setup:

- Creates database if necessary.
- Runs migrations.
- Manages backups pre-migration.

Methods:

- setup_db(app_config): setup a db with the given config.
"""

from contextlib import closing
import os
import sqlite3
import gzip
import shutil
from datetime import datetime
import glob

from .migrator import SqliteMigrator


class BackupManager:  # pylint: disable=too-few-public-methods
    """
    Creates db backups when needed, prunes old backups.
    """

    def __init__(self, file_to_backup, backup_directory, max_number_of_backups):
        self.file_to_backup = file_to_backup
        self.backup_directory = backup_directory
        self.max_number_of_backups = max_number_of_backups

    def do_backup(self, next_backup_datetime=None):
        """
        Perform the db file backup to backup_dir,
        pruning old backups.
        """

        if next_backup_datetime is None:
            now = datetime.now()
            next_backup_datetime = now.strftime("%Y%m%d-%H%M%S-%f")

        bname = os.path.basename(self.file_to_backup)
        backup_filename = f"{bname}.{next_backup_datetime}.gz"
        backup_path = os.path.join(self.backup_directory, backup_filename)

        os.makedirs(self.backup_directory, exist_ok=True)

        # Copy the file to the backup directory and gzip it
        with open(self.file_to_backup, "rb") as source_file, gzip.open(
            backup_path, "wb"
        ) as backup_file:
            shutil.copyfileobj(source_file, backup_file)
        assert os.path.exists(backup_path)

        # List all backup files in the directory, sorted by name.
        # Since this includes the timestamp, the oldest files will be
        # listed first.
        globname = f"{bname}.*.gz"
        globpath = os.path.join(self.backup_directory, globname)
        backup_files = glob.glob(globpath)
        backup_files.sort(key=os.path.basename)

        # Delete excess backup files if necessary
        while len(backup_files) > self.max_number_of_backups:
            file_to_delete = backup_files.pop(0)
            os.remove(file_to_delete)


class Setup:  # pylint: disable=too-few-public-methods
    """
    Main setup class, coordinates other classes.
    """

    def __init__(  # pylint: disable=too-many-arguments,too-many-positional-arguments
        self,
        db_filename: str,
        baseline_schema_file: str,
        backup_manager: BackupManager,
        migrator: SqliteMigrator,
        output_func=None,
    ):
        self._db_filename = db_filename
        self._baseline_schema_file = baseline_schema_file
        self._backup_mgr = backup_manager
        self._migrator = migrator
        self._output_func = output_func

    def setup(self):
        """
        Do database setup, making backup if necessary, running migrations.
        """
        new_db = False
        has_migrations = False

        if not os.path.exists(self._db_filename):
            new_db = True
            # Note openin a connection creates a db file,
            # so this has to be done after the existence
            # check.
            with closing(self._open_connection()) as conn:
                self._create_baseline(conn)

        with closing(self._open_connection()) as conn:
            has_migrations = self._migrator.has_migrations(conn)

        # Don't do a backup with an open connection ...
        # I don't _think_ it matters, but better to be safe.
        def null_print(s):  # pylint: disable=unused-argument
            pass

        outfunc = self._output_func or null_print
        if not new_db and has_migrations:
            outfunc("Creating backup before running migrations ...")
            self._backup_mgr.do_backup()
            outfunc("Done backup.")

        with closing(self._open_connection()) as conn:
            self._migrator.do_migration(conn)

    def _open_connection(self):
        """
        Get connection to db_filename.  Callers must close.
        """
        return sqlite3.connect(self._db_filename, detect_types=sqlite3.PARSE_DECLTYPES)

    def _create_baseline(self, conn):
        """
        Create baseline database.
        """
        b = self._baseline_schema_file
        with open(b, "r", encoding="utf8") as f:
            sql = f.read()
        conn.executescript(sql)


def _schema_dir():
    "Where schema files are found."
    thisdir = os.path.dirname(os.path.realpath(__file__))
    return os.path.join(thisdir, "..", "schema")


def _create_migrator():
    """
    Create SqliteMigrator with prod migration folders.
    """
    schema_dir = _schema_dir()
    migdir = os.path.join(schema_dir, "migrations")
    repeatable = os.path.join(schema_dir, "migrations_repeatable")
    return SqliteMigrator(migdir, repeatable)


def setup_db(app_config, output_func=None):
    """
    Main setup routine.
    """
    dbfile = app_config.dbfilename
    backup_dir = app_config.system_backup_path
    backup_count = 20  # Arbitrary
    bm = BackupManager(dbfile, backup_dir, backup_count)

    schema_dir = _schema_dir()
    baseline = os.path.join(schema_dir, "baseline.sql")

    migrator = _create_migrator()

    setup = Setup(dbfile, baseline, bm, migrator, output_func)
    setup.setup()



"""
Book Repository tests.

Tests lute.term.model.Book *domain* objects being saved
and retrieved from DB.
"""

import pytest

from lute.db import db
from lute.book.model import Book, Repository
from tests.dbasserts import assert_sql_result


@pytest.fixture(name="repo")
def fixture_repo():
    return Repository(db.session)


@pytest.fixture(name="new_book")
def fixture_book(english):
    """
    Term business object with some defaults,
    no tags or parents.
    """
    if english.id is None:
        db.session.add(english)
        db.session.commit()
    assert english.id is not None, "have english lang sanity check"
    b = Book()
    b.language_id = english.id
    b.title = "HELLO"
    b.text = "greeting"
    b.source_uri = "somesource"
    b.book_tags = ["tag1", "tag2"]
    return b


def test_save_new_book(app_context, new_book, repo):
    """
    Saving a simple Book object loads the database.
    """
    sql = "select BkTitle from books where BkTitle = 'HELLO'"
    assert_sql_result(sql, [], "empty table")

    b = repo.add(new_book)
    repo.commit()
    assert_sql_result(sql, ["HELLO"], "Saved")
    assert b.texts[0].text == "greeting"

    book = repo.load(b.id)
    assert book.title == new_book.title, "found book"
    assert book.book_tags == ["tag1", "tag2"], "tags filled"


def test_can_save_new_book_by_language_name(app_context, new_book, repo):
    """
    Can save a book with language name, useful for api access.
    """
    sql = "select BkTitle from books where BkTitle = 'HELLO'"
    assert_sql_result(sql, [], "empty table")

    new_book.language_id = None
    new_book.language_name = "English"
    b = repo.add(new_book)
    repo.commit()
    assert_sql_result(sql, ["HELLO"], "Saved")
    assert b.texts[0].text == "greeting"

    book = repo.load(b.id)
    assert book.title == new_book.title, "found book"
    assert book.book_tags == ["tag1", "tag2"], "tags filled"


def test_save_new_respects_book_words_per_page_count(app_context, new_book, repo):
    """
    Saving a simple Book object loads the database.
    """
    sql = "select BkTitle from books where BkTitle = 'HELLO'"
    assert_sql_result(sql, [], "empty table")

    new_book.max_page_tokens = 10
    new_book.text = (
        "One two three four. One two three four five six seven eight nine ten eleven."
    )
    b = repo.add(new_book)
    repo.commit()
    assert_sql_result(sql, ["HELLO"], "Saved")
    assert b.texts[0].text == "One two three four.", "page 1"
    assert (
        b.texts[1].text == "One two three four five six seven eight nine ten eleven."
    ), "page 2"

    book = repo.load(b.id)
    assert book.title == new_book.title, "found book"
    assert book.book_tags == ["tag1", "tag2"], "tags filled"


def test_get_tags(app_context, new_book, repo):
    "Helper method test."
    assert repo.get_book_tags() == [], "no tags yet"
    repo.add(new_book)
    repo.commit()
    assert repo.get_book_tags() == ["tag1", "tag2"], "no tags yet"


def test_save_update_existing(app_context, new_book, repo):
    "Update an existing book, tags are replaced."
    b = repo.add(new_book)
    repo.commit()
    assert_sql_result(
        "select BkTitle from books where BkTitle = 'HELLO'", ["HELLO"], "Saved"
    )

    book = repo.load(b.id)
    book.title = "CHANGED"
    book.book_tags = ["new1", "new2"]
    repo.add(book)
    repo.commit()

    assert_sql_result(
        "select BkTitle from books where BkTitle = 'CHANGED'", ["CHANGED"], "ok"
    )
    assert_sql_result(
        "select T2Text from tags2 order by T2Text", ["new1", "new2", "tag1", "tag2"]
    )


## Deletes.


def test_delete(app_context, new_book, repo):
    "Delete removes book, book must have an id."
    with pytest.raises(ValueError):
        repo.delete(new_book)

    b = repo.add(new_book)
    repo.commit()
    sql = "select BkTitle from books where BkTitle = 'HELLO'"
    assert_sql_result(sql, ["HELLO"], "Saved")

    foundbook = repo.load(b.id)
    repo.delete(foundbook)
    repo.commit()
    assert_sql_result(sql, [], "Deleted")



"""
Copy of main elements of main.py.

This still connects to the test database etc.
"""

import sys
import logging
from waitress import serve

if len(sys.argv) != 2:
    raise RuntimeError("have to pass port arg.")

# Hack the path, or python can't find the lute package when this is run
# from the root dir using "python -m this.module.name"
#
# pylint: disable=wrong-import-position
sys.path.append("..")
from lute.app_factory import create_app

logging.getLogger("waitress.queue").setLevel(logging.ERROR)

app = create_app()

port = int(sys.argv[1])
print(f"running at localhost:{port}", flush=True)
serve(app, host="0.0.0.0", port=port)



"""
TermTage mapping tests.
"""

from sqlalchemy import text
from lute.models.term import Term, TermTag
from lute.db import db
from tests.dbasserts import assert_record_count_equals

# from lute.termtag.routes import delete as route_delete


def test_deleting_termtag_removes_wordtags_table_record(empty_db, spanish):
    """
    Association record should be deleted if tag is deleted.

    Annoying test ... during unit testing, deleting TermTag entity
    causes the association table records wordtags to be deleted
    correctly, but during actual operation -- i.e., deletion of a
    TermTag through the UI -- the records aren't being deleted.
    Can't explain why, and I don't want to waste more time trying
    to figure it out.
    """

    tg = TermTag("tag")
    db.session.add(tg)
    db.session.commit()

    term = Term(spanish, "HOLA")
    term.add_term_tag(tg)
    db.session.add(term)
    db.session.commit()

    # The tag association for HOLA is getting deleted correctly
    # when the "tag" tag is deleted, which is odd because it's
    # not getting deleted when the action is called from the UI
    # (ref https://github.com/LuteOrg/lute-v3/issues/455).
    #
    # Trying adding another term, with directly inserting
    # the data in the table, to see if that is deleted correctly ...
    # ... and it is.
    perro = Term(spanish, "perro")
    db.session.add(perro)
    db.session.commit()

    sql = f"insert into wordtags (WtWoID, WtTgID) values ({perro.id}, {tg.id})"
    db.session.execute(text(sql))
    db.session.commit()

    # Trying loading data directly into the DB, so that the db session orm
    # isn't aware of a Term.  This too is deleted correctly.
    sql = f"""insert into words (WoLgID, WoText, WoTextLC, WoStatus)
      values ({spanish.id}, 'gato', 'gato', 1)"""
    db.session.execute(text(sql))
    db.session.commit()

    sql = f"insert into wordtags (WtWoID, WtTgID) values ({perro.id + 1}, {tg.id})"
    db.session.execute(text(sql))
    db.session.commit()

    sqltags = "select * from tags"
    assert_record_count_equals(sqltags, 1, "tag sanity check on save")

    sqlassoc = "select * from wordtags"
    assert_record_count_equals(sqlassoc, 3, "word tag associations exist")

    termtag = db.session.get(TermTag, tg.id)
    db.session.delete(termtag)
    db.session.commit()

    assert_record_count_equals(sqltags, 0, "tag removed")
    assert_record_count_equals(sqlassoc, 0, "associations removed")



"Theming routes."

from flask import Blueprint, Response, jsonify

from lute.themes.service import Service
from lute.models.repositories import UserSettingRepository
from lute.settings.current import current_settings
from lute.db import db

bp = Blueprint("themes", __name__, url_prefix="/theme")


@bp.route("/current", methods=["GET"])
def current_theme():
    "Return current css."
    service = Service(db.session)
    response = Response(service.get_current_css(), 200)
    response.content_type = "text/css; charset=utf-8"
    return response


@bp.route("/custom_styles", methods=["GET"])
def custom_styles():
    """
    Return the custom settings for inclusion in the base.html.
    """
    repo = UserSettingRepository(db.session)
    css = repo.get_value("custom_styles")
    response = Response(css, 200)
    response.content_type = "text/css; charset=utf-8"
    return response


@bp.route("/next", methods=["POST"])
def set_next_theme():
    "Go to next theme."
    service = Service(db.session)
    service.next_theme()
    return jsonify("ok")


@bp.route("/toggle_highlight", methods=["POST"])
def toggle_highlight():
    "Fix the highlight."
    new_setting = not current_settings["show_highlights"]
    repo = UserSettingRepository(db.session)
    repo.set_value("show_highlights", new_setting)
    db.session.commit()
    current_settings["show_highlights"] = new_setting
    return jsonify("ok")



"""
Show terms in datatables.
"""

from lute.utils.data_tables import DataTablesSqliteQuery, supported_parser_type_criteria


def get_data_tables_list(parameters, session):
    "Term json data for datatables."

    base_sql = """SELECT
    w.WoID as WoID, LgName, L.LgID as LgID, w.WoText as WoText, parents.parentlist as ParentText, w.WoTranslation,
    w.WoRomanization,
    replace(wi.WiSource, '.jpeg', '') as WiSource,
    ifnull(tags.taglist, '') as TagList,
    StText,
    StID,
    StAbbreviation,
    case w.WoSyncStatus when 1 then 'y' else '' end as SyncStatus,
    datetime(WoCreated, 'localtime') as WoCreated
    FROM
    words w
    INNER JOIN languages L on L.LgID = w.WoLgID
    INNER JOIN statuses S on S.StID = w.WoStatus
    LEFT OUTER JOIN (
      SELECT WpWoID as WoID, GROUP_CONCAT(PText, ', ') AS parentlist
      FROM
      (
        select WpWoID, WoText as PText
        from wordparents wp
        INNER JOIN words on WoID = WpParentWoID
        order by WoText
      ) parentssrc
      GROUP BY WpWoID
    ) AS parents on parents.WoID = w.WoID
    LEFT OUTER JOIN (
      SELECT WtWoID as WoID, GROUP_CONCAT(TgText, ', ') AS taglist
      FROM
      (
        select WtWoID, TgText
        from wordtags wt
        INNER JOIN tags t on t.TgID = wt.WtTgID
        order by TgText
      ) tagssrc
      GROUP BY WtWoID
    ) AS tags on tags.WoID = w.WoID
    LEFT OUTER JOIN wordimages wi on wi.WiWoID = w.WoID
    """

    typecrit = supported_parser_type_criteria()
    wheres = [f"L.LgParserType in ({typecrit})"]

    # Add "where" criteria for all the filters.

    # Have to check for 'null' for language filter.
    # A new user may filter the language when the demo data is loaded,
    # but on "wipe database" the filtLanguage value stored in localdata
    # may be invalid, resulting in the filtLanguage form control actually
    # sending the **string value** "null" here.
    # The other filter values don't change with the data,
    # so we don't need to check for null.
    # Tricky tricky.
    language_id = parameters["filtLanguage"]
    if language_id == "null" or language_id is None:
        language_id = "0"
    language_id = int(language_id)
    if language_id != 0:
        wheres.append(f"L.LgID == {language_id}")

    if parameters["filtParentsOnly"] == "true":
        wheres.append("parents.parentlist IS NULL")

    sql_age_calc = "cast(julianday('now') - julianday(w.wocreated) as int)"
    age_min = parameters["filtAgeMin"].strip()
    if age_min:
        wheres.append(f"{sql_age_calc} >= {int(age_min)}")
    age_max = parameters["filtAgeMax"].strip()
    if age_max:
        wheres.append(f"{sql_age_calc} <= {int(age_max)}")

    st_range = ["StID != 98"]
    status_min = int(parameters.get("filtStatusMin", "0"))
    status_max = int(parameters.get("filtStatusMax", "99"))
    st_range.append(f"StID >= {status_min}")
    st_range.append(f"StID <= {status_max}")
    st_where = " AND ".join(st_range)
    if parameters["filtIncludeIgnored"] == "true":
        st_where = f"({st_where} OR StID = 98)"
    wheres.append(st_where)

    # Phew.
    return DataTablesSqliteQuery.get_data(
        base_sql + " WHERE " + " AND ".join(wheres), parameters, session.connection()
    )



"""
Parsing using Jieba

The parser uses jieba to do parsing and pypinyin for character readings

Includes classes:

- MandarinParser

"""

import re
import os
from typing import List
import jieba
from pypinyin import pinyin
from lute.parse.base import ParsedToken, AbstractParser


class MandarinParser(AbstractParser):
    """
    A parser for Mandarin Chinese,
    using the jieba library for text segmentation.

    The user can add some exceptions to the "parsing_exceptions.txt"
    data file.
    """

    @classmethod
    def name(cls):
        return "Lute Mandarin Chinese"

    @classmethod
    def uses_data_directory(cls):
        "Uses the data_directory (defined in the AbstractParser)."
        return True

    @classmethod
    def parser_exceptions_file(cls):
        """Full path to an exceptions file (in the parser's
        data_directory) to indicate which terms should be broken up
        differently than what jieba suggests.  For example, jieba
        parses "清华大学" as a single token; however the user can
        specify different parsing for this group:

        "清华,大学" says "parse 清华大学 into two tokens, 清华/大学."
        "清,华,大学" says "parse 清华大学 into three tokens, 清/华/大学."

        Each rule is placed on a separate line in the
        parser_exceptions file, e.g, the following file content
        defines two rules:

        清华,大学
        学,华,大
        """
        return os.path.join(cls.data_directory, "parser_exceptions.txt")

    @classmethod
    def init_data_directory(cls):
        "Set up necessary files."
        fp = cls.parser_exceptions_file()
        if not os.path.exists(fp):
            with open(fp, "w", encoding="utf8") as f:
                f.write("# Parsing exceptions.\n")
                f.write("# Place each rule on a separate line. e.g.:\n")
                f.write("# 清华,大学\n")
                f.write("# Lines preceded with # are ignored.\n")

    @classmethod
    def _build_parser_exceptions_map(cls):
        "Convert exceptions file to map."
        if cls.data_directory is None:
            return {}

        ret = {}
        with open(cls.parser_exceptions_file(), "r", encoding="utf8") as f:
            for line in f:
                stripped_line = line.strip()
                if stripped_line.startswith("#"):
                    continue
                parts = [p.strip() for p in stripped_line.split(",")]
                orig_token = "".join(parts)
                ret[orig_token] = parts
        return ret

    def _reparse_with_exceptions_map(self, original_token, exceptions_map):
        "Check the token s against the map, break down further if needed."

        # pylint: disable=dangerous-default-value
        def _get_mapped(tok, accum=[]):
            parts = exceptions_map.get(tok)
            if parts is None or len(parts) == 1:
                accum.append(tok)
            else:
                for p in parts:
                    _get_mapped(p, accum)
            return accum

        return _get_mapped(original_token)

    def get_parsed_tokens(self, text: str, language) -> List[ParsedToken]:
        """
        Returns ParsedToken array for given language.
        """

        exceptions_map = self._build_parser_exceptions_map()

        # Ensure standard carriage returns so that paragraph
        # markers are used correctly.  Lute uses paragraph markers
        # for rendering.
        text = text.replace("\r\n", "\n")

        words = list(jieba.cut(text))
        tokens = []
        pattern = f"[{language.word_characters}]"
        for word in words:
            is_word_char = re.match(pattern, word) is not None
            is_end_of_sentence = word in language.regexp_split_sentences
            if word == "\n":
                word = "¶"
            if word == "¶":
                is_word_char = False
                is_end_of_sentence = True
            parts = self._reparse_with_exceptions_map(word, exceptions_map)
            for p in parts:
                t = ParsedToken(p, is_word_char, is_end_of_sentence)
                tokens.append(t)
        return tokens

    def get_reading(self, text: str):
        """
        Get the pinyin for the given text.
        Returns None if the text is all Chinese characters, or the pinyin
        doesn't add value (same as text).
        """
        # Use pypinyin to get the pinyin of the text
        pinyin_list = pinyin(text)
        # Flatten the list of lists to a single list
        pinyin_list = (item for sublist in pinyin_list for item in sublist)
        # Join the pinyin into a single string
        ret = " ".join(pinyin_list)
        if ret in ("", text):
            return None
        return ret



"""
Version info.

Lute follows the version numbers at
https://packaging.python.org/en/latest/specifications/version-specifiers/#version-specifiers

e.g.

3.0.0a1.dev1
3.0.0a1
3.0.0b1
3.0.0

The version needs to be included in Lute itself, because Lute displays
it in the application version screen.

Flit pulls into the pyproject.toml using "dynamic".
"""

__version__ = "3.8.1b1"



"""
Step defs for term_rendering.feature.
"""
# pylint: disable=missing-function-docstring

from pytest_bdd import given, then, scenarios, parsers

from lute.db import db
from lute.models.language import Language
from lute.language.service import Service as LanguageService
from lute.term.model import Repository
from lute.read.render.service import Service as RenderService
from lute.read.service import Service

from tests.utils import add_terms, make_text
from tests.dbasserts import assert_sql_result


# The current language being used.
language = None

# The Text object
text = None

scenarios("rendering.feature")


@given("demo data")
def given_demo_data(app_context):
    "Calling app_context loads the demo data."
    # TODO remove this


@given(parsers.parse("language {langname}"))
def given_lang(langname):
    svc = LanguageService(db.session)
    svc.load_language_def(langname)
    global language  # pylint: disable=global-statement
    lang = db.session.query(Language).filter(Language.name == langname).first()
    assert lang.name == langname, "sanity check"
    language = lang


@given(parsers.parse("terms:\n{content}"))
def given_terms(content):
    terms = content.split("\n")
    add_terms(language, terms)


@given(parsers.parse('term "{content}" with status {status} and parent "{parenttext}"'))
def given_term_with_status_and_parent(content, status, parenttext):
    r = Repository(db.session)
    t = r.find_or_new(language.id, content)
    t.status = int(status)
    t.parents.append(parenttext)
    r.add(t)
    r.commit()


@given(parsers.parse('term "{content}" with status {status}'))
def given_term_with_status(content, status):
    r = Repository(db.session)
    t = r.find_or_new(language.id, content)
    t.status = int(status)
    r.add(t)
    r.commit()


@given(parsers.parse("text:\n{content}"))
def given_text(content):
    global text  # pylint: disable=global-statement
    text = make_text("test", content, language)
    db.session.add(text)
    db.session.commit()


@given("all unknowns are set to known")
def set_to_known():
    service = Service(db.session)
    service.set_unknowns_to_known(text)


@given(parsers.parse("bulk status {newstatus} update for terms:\n{terms}"))
def update_status(newstatus, terms):
    service = Service(db.session)
    service.bulk_status_update(text, terms.split("\n"), int(newstatus))


def _assert_stringized_equals(stringizer, joiner, expected):
    """
    Get paragraphs and stringize all textitems,
    join and assert equals expected.
    """
    rs = RenderService(db.session)
    paras = rs.get_paragraphs(text.text, text.book.language)
    # print("TOKENS", flush=True)
    # print(paras, flush=True)
    ret = []
    for p in paras:
        tis = [t for s in p for t in s]
        ss = [stringizer(ti) for ti in tis]
        ret.append(joiner.join(ss))
    actual = "/<PARA>/".join(ret)

    # print("", flush=True)
    # print(expected, flush=True)
    expected = "/<PARA>/".join(expected.split("\n"))
    assert actual == expected


@then(parsers.parse("rendered should be:\n{expected}"))
def then_rendered_should_be(expected):
    """
    Renders /term(status)/ /term/ /term/, compares with expected.
    """

    def stringize(ti):
        zws = "\u200B"
        status = ""
        if ti.wo_status not in [None, 0]:
            status = f"({ti.wo_status})"
        return ti.display_text.replace(zws, "") + status

    _assert_stringized_equals(stringize, "/", expected)


@then(parsers.parse("known-only rendered should be:\n{expected}"))
def known_only_rendered_should_be(expected):
    def stringize(ti):
        s = ti.display_text
        if ti.wo_status not in [None, 0]:
            s = f"[[{s}]]"
        zws = "\u200B"
        return s.replace(zws, "")

    _assert_stringized_equals(stringize, "", expected)


@then(parsers.parse("words table should contain:\n{text_lc_content}"))
def then_words_table_contains_WoTextLC(text_lc_content):
    expected = text_lc_content.split("\n")
    sql = "select WoTextLC from words order by WoTextLC"
    assert_sql_result(sql, expected)



"""
Term mapping tests.
"""

import html
from bs4 import BeautifulSoup
from lute.models.term import Term, TermTag
from lute.models.repositories import UserSettingRepository
from lute.db import db


def test_smoke_popup_response(client, empty_db, spanish):
    "Misc data check - parent and tags are saved."
    term = Term(spanish, "un gatito")
    term.translation = "t_trans\nt_extra"
    term.romanization = "t_rom"
    term.set_flash_message("hello")
    term.add_term_tag(TermTag("t_tag"))
    term.set_current_image("blah.jpg")

    parent = Term(spanish, "un gato")
    parent.translation = "p_trans\np_extra"
    parent.romanization = "p_rom"
    parent.add_term_tag(TermTag("p_tag"))
    parent.set_current_image("parentblah.jpg")

    component = Term(spanish, "gatito")
    component.translation = "c_trans\nc_extra"
    component.romanization = "c_rom"
    component.add_term_tag(TermTag("c_tag"))

    term.add_parent(parent)
    db.session.add(term)
    db.session.add(component)
    db.session.commit()

    term.set_flash_message("hello")

    db.session.add(term)
    db.session.add(parent)
    db.session.add(component)
    db.session.commit()

    def _get_pretty_response(term_id):
        "Response for term popup."
        response = client.get(f"/read/termpopup/{term_id}")
        decoded_response = response.data.decode("utf-8")
        unescaped_response = html.unescape(decoded_response)
        soup = BeautifulSoup(unescaped_response, "html.parser")
        pretty_response = soup.prettify()
        return pretty_response

    pretty_response = _get_pretty_response(term.id)
    print(pretty_response, flush=True)
    for t in ["t", "p", "c"]:
        for part in ["trans", "rom", "tag"]:
            s = f"{t}_{part}"
            assert s in pretty_response, s

    us_repo = UserSettingRepository(db.session)
    us_repo.set_value("term_popup_show_components", False)
    db.session.commit()
    pretty_response = _get_pretty_response(term.id)
    print(pretty_response, flush=True)
    for part in ["trans", "rom", "tag"]:
        s = f"c_{part}"
        assert s not in pretty_response, s



"""
Settings form.
"""

import os
from flask_wtf import FlaskForm
from wtforms import BooleanField, StringField, IntegerField, TextAreaField, SelectField
from wtforms.validators import InputRequired, NumberRange
from wtforms import ValidationError


class UserSettingsForm(FlaskForm):
    """
    Settings.

    Note the field names here must match the keys in the settings table.
    """

    backup_enabled = BooleanField("Backup Enabled")
    backup_dir = StringField("Backup directory")
    backup_auto = BooleanField("Run backups automatically (daily)")
    backup_warn = BooleanField("Warn if backup hasn't run in a week")
    backup_count = IntegerField(
        "Retain backup count",
        validators=[InputRequired(), NumberRange(min=1)],
        render_kw={
            "title": "Count of zipfiles to retain, oldest files are deleted first"
        },
    )

    current_theme = SelectField("Theme")
    custom_styles = TextAreaField("Custom styles")
    show_highlights = BooleanField("Highlight terms by status")

    open_popup_in_new_tab = BooleanField("Open popup in new tab")
    stop_audio_on_term_form_open = BooleanField("Stop audio on term form open")
    stats_calc_sample_size = IntegerField(
        "Book stats page sample size",
        validators=[InputRequired(), NumberRange(min=1, max=500)],
        render_kw={"title": "Number of pages to use for book stats calculation."},
    )

    term_popup_promote_parent_translation = BooleanField(
        "Promote parent translation to term translation if possible"
    )
    term_popup_show_components = BooleanField("Show component terms")

    mecab_path = StringField("MECAB_PATH environment variable")
    reading_choices = [
        ("katakana", "Katakana"),
        ("hiragana", "Hiragana"),
        ("alphabet", "Romaji"),
    ]
    japanese_reading = SelectField("Pronunciation characters", choices=reading_choices)

    def validate_backup_dir(self, field):
        "Field must be set if enabled."
        if self.backup_enabled.data is False:
            return
        v = field.data
        if (v or "").strip() == "":
            raise ValidationError("Backup directory required")

        abspath = os.path.abspath(v)
        if v != abspath:
            msg = f'Backup dir must be absolute path.  Did you mean "{abspath}"?'
            raise ValidationError(msg)
        if not os.path.exists(v):
            raise ValidationError(f'Directory "{v}" does not exist.')
        if not os.path.isdir(v):
            raise ValidationError(f'"{v}" is not a directory.')


class UserShortcutsForm(FlaskForm):
    """
    Shortcuts form.

    The route manages getting and storing the settings
    from the db, as there's a variable number of settings,
    and it's easier to just work with the data directly
    rather than trying to create a variable number of fields.

    I'm only using this form to get the validate_on_submit()!
    There's likely a better way to do this.
    """



"""
TermTage mapping tests.
"""

from lute.models.term import Term, TermTag
from lute.db import db
from lute.termtag.routes import delete as route_delete
from tests.dbasserts import assert_record_count_equals


def test_deleting_termtag_removes_wordtags_table_record(empty_db, spanish):
    "Association record should be deleted if tag is deleted."

    tg = TermTag("tag")
    db.session.add(tg)
    db.session.commit()

    term = Term(spanish, "HOLA")
    term.add_term_tag(tg)
    db.session.add(term)
    db.session.commit()

    sqlterms = "select * from words"
    assert_record_count_equals(sqlterms, 1, "term sanity check on save")

    sqltags = "select * from tags"
    assert_record_count_equals(sqltags, 1, "tag sanity check on save")

    sqlassoc = "select * from wordtags"
    assert_record_count_equals(sqlassoc, 1, "word tag associations exist")

    route_delete(tg.id)

    assert_record_count_equals(sqlterms, 1, "term stays")
    assert_record_count_equals(sqltags, 0, "tag removed")
    assert_record_count_equals(sqlassoc, 0, "associations removed")



"""
Db and image backup.
"""

import os
import re
import shutil
import gzip
from datetime import datetime
import time
from typing import List, Union

from lute.models.repositories import UserSettingRepository
from lute.models.book import Book
from lute.models.term import Term


class BackupException(Exception):
    """
    Raised if export bombs for some reason.
    """


class DatabaseBackupFile:
    """
    A representation of a lute backup file to hold metadata attributes.
    """

    def __init__(self, filepath: Union[str, os.PathLike]):
        if not os.path.exists(filepath):
            raise BackupException(f"No backup file at {filepath}.")

        name = os.path.basename(filepath)
        if not re.match(r"(manual_)?lute_backup_", name):
            raise BackupException(f"Not a valid lute database backup at {filepath}.")

        self.filepath = filepath
        self.name = name
        self.is_manual = self.name.startswith("manual_")

    def __lt__(self, other):
        return self.last_modified < other.last_modified

    @property
    def last_modified(self) -> datetime:
        return datetime.fromtimestamp(os.path.getmtime(self.filepath)).astimezone()

    @property
    def size_bytes(self) -> int:
        return os.path.getsize(self.filepath)

    @property
    def size(self) -> str:
        """
        A human-readable string representation of the size of the file.

        Eg.
        1746 bytes
        4 kB
        27 MB
        """
        s = self.size_bytes
        if s >= 1e9:
            return f"{round(s * 1e-9)} GB"
        if s >= 1e6:
            return f"{round(s * 1e-6)} MB"
        if s >= 1e3:
            return f"{round(s * 1e-3)} KB"
        return f"{s} bytes"


class Service:
    "Service."

    def __init__(self, session):
        self.session = session

    def create_backup(self, app_config, settings, is_manual=False, suffix=None):
        """
        Create backup using current app config, settings.

        is_manual is True if this is a user-triggered manual
        backup, otherwise is False.

        suffix can be specified for test.

        settings are from BackupSettings.
          - backup_enabled
          - backup_dir
          - backup_auto
          - backup_warn
          - backup_count
          - last_backup_datetime
        """
        if not os.path.exists(settings.backup_dir):
            raise BackupException("Missing directory " + settings.backup_dir)

        # def _print_now(msg):
        #     "Timing helper for when implement audio backup."
        #     now = datetime.now().strftime("%H-%M-%S")
        #     print(f"{now} - {msg}", flush=True)

        self._mirror_images_dir(app_config.userimagespath, settings.backup_dir)

        prefix = "manual_" if is_manual else ""
        if suffix is None:
            suffix = datetime.now().strftime("%Y-%m-%d_%H%M%S")
        fname = f"{prefix}lute_backup_{suffix}.db"
        backupfile = os.path.join(settings.backup_dir, fname)

        f = self._create_db_backup(app_config.dbfilename, backupfile)
        self._remove_excess_backups(settings.backup_count, settings.backup_dir)
        return f

    def should_run_auto_backup(self, backup_settings):
        """
        True (if applicable) if last backup was old.
        """
        bs = backup_settings
        if bs.backup_enabled is False or bs.backup_auto is False:
            return False

        last = bs.last_backup_datetime
        if last is None:
            return True

        curr = int(time.time())
        diff = curr - last
        return diff > 24 * 60 * 60

    def backup_warning(self, backup_settings):
        "Get warning if needed."
        if not backup_settings.backup_warn:
            return ""

        have_books = self.session.query(self.session.query(Book).exists()).scalar()
        have_terms = self.session.query(self.session.query(Term).exists()).scalar()
        if have_books is False and have_terms is False:
            return ""

        last = backup_settings.last_backup_datetime
        if last is None:
            return "Never backed up."

        curr = int(time.time())
        diff = curr - last
        old_backup_msg = "Last backup was more than 1 week ago."
        if diff > 7 * 24 * 60 * 60:
            return old_backup_msg

        return ""

    def _create_db_backup(self, dbfilename, backupfile):
        "Make a backup."
        shutil.copy(dbfilename, backupfile)
        f = f"{backupfile}.gz"
        with open(backupfile, "rb") as in_file, gzip.open(
            f, "wb", compresslevel=4
        ) as out_file:
            shutil.copyfileobj(in_file, out_file)
        os.remove(backupfile)
        r = UserSettingRepository(self.session)
        r.set_last_backup_datetime(int(time.time()))
        return f

    def skip_this_backup(self):
        "Set the last backup time to today."
        r = UserSettingRepository(self.session)
        r.set_last_backup_datetime(int(time.time()))

    def _remove_excess_backups(self, count, outdir):
        "Remove old backups."
        files = [f for f in self.list_backups(outdir) if not f.is_manual]
        files.sort(reverse=True)
        to_remove = files[count:]
        for f in to_remove:
            os.remove(f.filepath)

    def _mirror_images_dir(self, userimagespath, outdir):
        "Copy the images to backup."
        target_dir = os.path.join(outdir, "userimages_backup")
        target_dir = os.path.abspath(target_dir)
        if not os.path.exists(target_dir):
            os.mkdir(target_dir)
        shutil.copytree(userimagespath, target_dir, dirs_exist_ok=True)

    def list_backups(self, outdir) -> List[DatabaseBackupFile]:
        "List all backup files."
        return [
            DatabaseBackupFile(os.path.join(outdir, f))
            for f in os.listdir(outdir)
            if re.match(r"(manual_)?lute_backup_", f)
        ]



"""
Common classes use for all parsing.
"""

from abc import ABC, abstractmethod
from typing import List


class ParsedToken:
    """
    A single parsed token from an input text.

    As tokens are created, the class counters
    (starting with cls_) are assigned to the ParsedToken
    and then incremented appropriately.
    """

    # Class counters.
    cls_sentence_number = 0
    cls_order = 0

    @classmethod
    def reset_counters(cls):
        """
        Reset all the counters.
        """
        ParsedToken.cls_sentence_number = 0
        ParsedToken.cls_order = 0

    def __init__(self, token: str, is_word: bool, is_end_of_sentence: bool = False):
        self.token = token
        self.is_word = is_word
        self.is_end_of_sentence = is_end_of_sentence

        ParsedToken.cls_order += 1
        self.order = ParsedToken.cls_order

        self.sentence_number = ParsedToken.cls_sentence_number

        # Increment counters after the TextToken has been
        # completed, so that it belongs to the correct
        # sentence.
        if self.is_end_of_sentence:
            ParsedToken.cls_sentence_number += 1

    def __repr__(self):
        attrs = [
            f"word: {self.is_word}",
            f"eos: {self.is_end_of_sentence}",
            # f"sent: {self.sentence_number}",
        ]
        attrs = ", ".join(attrs)
        return f'<"{self.token}" ({attrs})>'


class SentenceGroupIterator:
    """
    An iterator of ParsedTokens that groups them by sentence, up
    to a maximum number of tokens.
    """

    def __init__(self, tokens, maxcount=500):
        self.tokens = tokens
        self.maxcount = maxcount
        self.currpos = 0

    def count(self):
        """
        Get count of groups that will be returned.
        """
        old_currpos = self.currpos
        c = 0
        while self.next():
            c += 1
        self.currpos = old_currpos
        return c

    def next(self):
        """
        Get next sentence group.
        """
        if self.currpos >= len(self.tokens):
            return False

        curr_tok_count = 0
        last_eos = -1
        i = self.currpos

        while (curr_tok_count <= self.maxcount or last_eos == -1) and i < len(
            self.tokens
        ):
            tok = self.tokens[i]
            if tok.is_end_of_sentence == 1:
                last_eos = i
            if tok.is_word == 1:
                curr_tok_count += 1
            i += 1

        if curr_tok_count <= self.maxcount or last_eos == -1:
            ret = self.tokens[self.currpos : i]
            self.currpos = i + 1
        else:
            ret = self.tokens[self.currpos : last_eos + 1]
            self.currpos = last_eos + 1

        return ret


class AbstractParser(ABC):
    """
    Abstract parser, inherited from by all parsers.

    Attributes:
        data_directory: Optional.  A full path to a
            directory that the parser uses.  Should be
            initialized with init_data_directory().
    """

    data_directory = None

    @classmethod
    def uses_data_directory(cls):
        "True if the parser needs user-supplied data."
        return False

    @classmethod
    def init_data_directory(cls):
        """
        Initialize the data_directory if needed.  Not
        necessary for all parsers.
        """
        return

    @classmethod
    def is_supported(cls):
        """
        True if the parser will work on the current system.

        Some parsers, such as Japanese, require external
        components to be present and configured.  If missing,
        this should return False.
        """
        return True

    @classmethod
    @abstractmethod
    def name(cls):
        """
        Parser name, for displaying in UI.
        """

    @abstractmethod
    def get_parsed_tokens(self, text: str, language) -> List:
        """
        Get an array of ParsedTokens from the input text for the given language.
        """

    def get_reading(self, text: str):  # pylint: disable=unused-argument
        """
        Get the pronunciation for the given text.  For most
        languages, this can't be automated.
        """
        return None

    def get_lowercase(self, text: str):
        """
        Return the lowcase text.

        Most languages can use the built-in lowercase operation,
        but some (like Turkish) need special handling.
        """
        return text.lower()



"""
Language mapping tests.
"""

import json
from lute.models.language import Language, LanguageDictionary
from lute.models.repositories import LanguageRepository
from lute.read.service import Service as ReadService
from lute.db import db
from tests.dbasserts import assert_sql_result, assert_record_count_equals
from tests.utils import make_text, add_terms


def test_save_new_language(empty_db):
    """
    Check language mappings and defaults.
    """
    sql = """select LgName, LgRightToLeft,
    LgShowRomanization, LgRegexpSplitSentences
    from languages"""
    assert_sql_result(sql, [], "empty table")

    lang = Language()
    lang.name = "abc"

    db.session.add(lang)
    db.session.commit()
    assert_sql_result(sql, ["abc; 0; 0; .!?"], "have language, defaults as expected")

    lang.right_to_left = True
    db.session.add(lang)
    db.session.commit()
    assert_sql_result(sql, ["abc; 1; 0; .!?"], "rtl is True")

    retrieved = db.session.query(Language).filter(Language.name == "abc").first()
    # print(retrieved)
    assert retrieved.name == "abc"
    assert retrieved.right_to_left is True, "retrieved is RTL"
    assert retrieved.show_romanization is False, "retrieved no roman"


def test_language_dictionaries_smoke_test(empty_db):
    "Smoke test for new dictionary structure."
    lang = Language()
    lang.name = "abc"

    ld = LanguageDictionary()
    ld.usefor = "terms"
    ld.dicttype = "embeddedhtml"
    ld.dicturi = "1?[LUTE]"
    ld.sort_order = 1
    lang.dictionaries.append(ld)
    ld2 = LanguageDictionary()
    ld2.usefor = "terms"
    ld2.dicttype = "popuphtml"
    ld2.dicturi = "2?[LUTE]"
    ld2.sort_order = 2
    lang.dictionaries.append(ld2)

    ld3 = LanguageDictionary()
    ld3.usefor = "sentences"
    ld3.dicttype = "popuphtml"
    ld3.dicturi = "3?[LUTE]"
    ld3.sort_order = 3
    lang.dictionaries.append(ld3)

    db.session.add(lang)
    db.session.commit()

    sqldicts = """select LgName, LdUseFor, LdType, LdDictURI
    from languages
    inner join languagedicts on LdLgID = LgID
    order by LdSortOrder"""
    assert_sql_result(
        sqldicts,
        [
            "abc; terms; embeddedhtml; 1?[LUTE]",
            "abc; terms; popuphtml; 2?[LUTE]",
            "abc; sentences; popuphtml; 3?[LUTE]",
        ],
        "dict saved",
    )

    retrieved = db.session.query(Language).filter(Language.name == "abc").first()
    assert len(retrieved.dictionaries) == 3, "have dicts"
    ld = retrieved.dictionaries[0]
    assert ld.dicttype == "embeddedhtml", "type"
    assert ld.dicturi == "1?[LUTE]", "uri"

    exp = """{"1": {"term": ["1?[LUTE]", "*2?[LUTE]"], "sentence": ["*3?[LUTE]"]}}"""
    repo = LanguageRepository(db.session)
    dicts = repo.all_dictionaries()
    assert json.dumps(dicts) == exp


def test_delete_language_removes_book_and_terms(app_context, spanish):
    """
    Test HACKY LanguageRepository.delete() method to ensure deletes cascade.
    """
    add_terms(spanish, ["gato", "perro"])
    t = make_text("hola", "Hola amigo", spanish)
    db.session.add(t)

    ld = LanguageDictionary()
    ld.usefor = "terms"
    ld.dicttype = "embeddedhtml"
    ld.dicturi = "something?[LUTE]"
    ld.sort_order = 1
    spanish.dictionaries.append(ld)
    db.session.add(spanish)

    db.session.commit()

    svc = ReadService(db.session)
    svc.mark_page_read(t.book.id, 1, False)

    sqlterms = "select WoText from words order by WoText"
    sqlbook = "select BkTitle from books where BkTitle = 'hola'"
    sqldict = "select LdDictURI from languagedicts where LdDictURI = 'something?[LUTE]'"
    assert_sql_result(sqlterms, ["gato", "perro"], "initial terms")
    assert_sql_result(sqlbook, ["hola"], "initial book")
    assert_sql_result(sqldict, ["something?[LUTE]"], "dict")
    assert_record_count_equals("select * from wordsread", 1, "saved")

    db.session.delete(spanish)
    db.session.commit()

    assert_sql_result(sqlbook, [], "book deleted")
    assert_sql_result(sqlterms, [], "terms deleted")
    assert_sql_result(sqldict, [], "dicts deleted")
    assert_record_count_equals("select * from wordsread", 0, "deleted")



"""
Getting and saving bing image search results.
"""

import os
import re
import urllib.request
from flask import Blueprint, request, Response, render_template, jsonify, current_app


bp = Blueprint("bing", __name__, url_prefix="/bing")


@bp.route("/search/<int:langid>/<string:text>/<string:searchstring>", methods=["GET"])
def bing_search(langid, text, searchstring):
    "Do an image search."

    # Searching for images slows acceptance tests.  If NO_BING_IMAGES
    # environment setting, don't do a search.
    if "NO_BING_IMAGES" in os.environ:
        return render_template(
            "imagesearch/index.html", langid=langid, text=text, images=[]
        )

    # dump("searching for " + text + " in " + language.getLgName())
    search = urllib.parse.quote(text)
    params = searchstring.replace("[LUTE]", search)
    params = params.replace("###", search)  # TODO remove_old_###_placeholder: remove
    url = "https://www.bing.com/images/search?" + params
    content = ""
    error_msg = ""
    try:
        with urllib.request.urlopen(url) as s:
            content = s.read().decode("utf-8")
    except urllib.error.URLError as e:
        content = ""
        error_msg = e.reason

    # Sample data returned by bing image search:
    # <img class="mimg vimgld" ... data-src="https:// ...">
    # or
    # <img class="mimg rms_img" ... src="https://tse4.mm.bing ..." >

    def is_search_img(img):
        return not ('src="/' in img) and ("rms_img" in img or "vimgld" in img)

    def build_struct(image):
        src = "missing"
        normalized_source = image.replace("data-src=", "src=")
        m = re.search(r'src="(.*?)"', normalized_source)
        if m:
            src = m.group(1)
        return {"html": image, "src": src}

    raw_images = list(re.findall(r"(<img .*?>)", content, re.I))

    images = [build_struct(i) for i in raw_images if is_search_img(i)]

    # Reduce image load count so we don't kill subpage loading.
    # Also bing seems to throttle images if the count is higher (??).
    images = images[:25]

    return render_template(
        "imagesearch/index.html",
        langid=langid,
        text=text,
        images=images,
        error_message=error_msg,
    )


def _get_dir_and_filename(langid, text):
    "Make a directory if needed, return [dir, filename]"
    datapath = current_app.config["DATAPATH"]
    image_dir = os.path.join(datapath, "userimages", langid)
    if not os.path.exists(image_dir):
        os.makedirs(image_dir)
    filename = re.sub(r"\s+", "_", text) + ".jpeg"
    return [image_dir, filename]


@bp.route("/save", methods=["POST"])
def bing_save():
    """
    Save the posted image data to DATAPATH/userimages,
    returning the filename.
    """
    src = request.form["src"]
    text = request.form["text"]
    langid = request.form["langid"]

    imgdir, filename = _get_dir_and_filename(langid, text)
    destfile = os.path.join(imgdir, filename)
    with urllib.request.urlopen(src) as response, open(destfile, "wb") as out_file:
        out_file.write(response.read())

    # This is the format of legacy Lute v2 data.
    image_url = f"/userimages/{langid}/{filename}"
    return jsonify({"filename": image_url})


@bp.route("/manual_image_post", methods=["POST"])
def manual_image_post():
    """
    For manual posts of images (not bing image clicks).
    Save the posted image data to DATAPATH/userimages,
    returning the filename.
    """
    text = request.form["text"]
    langid = request.form["langid"]

    if "manual_image_file" not in request.files:
        return Response("No file part in request", status=400)

    f = request.files["manual_image_file"]
    if f.filename == "":
        return Response("No selected file", status=400)

    imgdir, filename = _get_dir_and_filename(langid, text)
    destfile = os.path.join(imgdir, filename)
    f.save(destfile)

    # This is the format of legacy Lute v2 data.
    image_url = f"/userimages/{langid}/{filename}"
    return jsonify({"filename": image_url})



"""
Parsing for single-character languages.

The parser uses some Language settings (e.g., word characters) to
perform the actual parsing.

Includes classes:

- ClassicalChineseParser

"""

import re
from typing import List
from lute.parse.base import ParsedToken, AbstractParser


class ClassicalChineseParser(AbstractParser):
    """
    A general parser for space-delimited languages,
    such as English, French, Spanish ... etc.
    """

    @classmethod
    def name(cls):
        return "Classical Chinese"

    def get_parsed_tokens(self, text: str, language) -> List[ParsedToken]:
        """
        Returns ParsedToken array for given language.
        """
        text = re.sub(r"[ \t]+", "", text)

        replacements = language.character_substitutions.split("|")
        for replacement in replacements:
            fromto = replacement.strip().split("=")
            if len(fromto) >= 2:
                rfrom = fromto[0].strip()
                rto = fromto[1].strip()
                text = text.replace(rfrom, rto)

        text = text.replace("\r\n", "\n")
        text = text.replace("{", "[")
        text = text.replace("}", "]")
        text = text.replace("\n", "¶")
        text = text.strip()

        tokens = []
        pattern = f"[{language.word_characters}]"
        for char in text:
            is_word_char = re.match(pattern, char) is not None
            is_end_of_sentence = char in language.regexp_split_sentences
            if char == "¶":
                is_end_of_sentence = True
            p = ParsedToken(char, is_word_char, is_end_of_sentence)
            tokens.append(p)

        return tokens



"""
Current user settings stored in UserSettings.

Storing a global dict to allow for db-less access, they're
global settings, after all.

They're written to at load (or when the settings change).
"""

from lute.models.setting import UserSetting

# The current user settings, key/value dict.
current_settings = {}

# Current user hotkey mappings, mapping to mapping_name dict.
current_hotkeys = {}


def refresh_global_settings(session):
    "Refresh all settings dictionary."
    # Have to reload to not mess up any references
    # (e.g. during testing).
    current_settings.clear()
    current_hotkeys.clear()

    settings = session.query(UserSetting).all()
    for s in settings:
        current_settings[s.key] = s.value

    hotkeys = [
        s for s in settings if s.key.startswith("hotkey_") and (s.value or "") != ""
    ]
    for h in hotkeys:
        current_hotkeys[h.value] = h.key

    # Convert some ints into bools.
    boolkeys = [
        "open_popup_in_new_tab",
        "stop_audio_on_term_form_open",
        "show_highlights",
        "term_popup_promote_parent_translation",
        "term_popup_show_components",
    ]
    for k in boolkeys:
        current_settings[k] = current_settings[k] == "1"



"""
Developer entry point.  Runs lute on a dev server.

This script _always_ uses the config at /lute/config/config.yml.

You can run with:

inv start
python -m devstart

If you want to run this with "python", then for some _extremely odd_
reason, you must run this with the full path to the file.
Ref https://stackoverflow.com/questions/37650208/flask-cant-find-app-file

e.g.:

python /Users/jeff/Documents/Projects/lute_v3/devstart.py
python `pwd`/devstart.py
"""

import os
import argparse
import logging
from lute import __version__
from lute.app_factory import create_app, data_initialization
from lute.config.app_config import AppConfig
from lute.db import db

log = logging.getLogger("werkzeug")
log.setLevel(logging.ERROR)


def start(port):
    """
    Start the dev server with reloads on port.
    """

    def dev_print(s):
        "Print info on first load only."
        if os.environ.get("WERKZEUG_RUN_MAIN") == "true":
            # https://stackoverflow.com/questions/25504149/
            #  why-does-running-the-flask-dev-server-run-itself-twice
            # Reloading, do nothing.
            return
        print(s, flush=True)

    config_file = AppConfig.default_config_filename()
    dev_print("")
    app = create_app(config_file, output_func=dev_print)
    with app.app_context():
        data_initialization(db.session, dev_print)

    ac = AppConfig(config_file)
    dev_print(f"\nversion {__version__}")
    dev_print(f"db name: {ac.dbname}")
    dev_print(f"data: {ac.datapath}")
    dev_print(f"Running at: http://localhost:{port}\n")

    app.run(debug=True, port=port)


if __name__ == "__main__":
    parser = argparse.ArgumentParser(description="Start dev server lute.")
    parser.add_argument(
        "--port", type=int, default=5001, help="Port number (default: 5001)"
    )
    args = parser.parse_args()
    start(args.port)



"""
Show books in datatables.
"""

from lute.utils.data_tables import DataTablesSqliteQuery, supported_parser_type_criteria


def get_data_tables_list(parameters, is_archived, session):
    "Book json data for datatables."
    archived = "true" if is_archived else "false"

    base_sql = f"""
    SELECT
        b.BkID As BkID,
        LgName,
        BkTitle,
        case when currtext.TxID is null then 1 else currtext.TxOrder end as PageNum,
        textcounts.pagecount AS PageCount,
        BkArchived,
        tags.taglist AS TagList,
        textcounts.wc AS WordCount,
        c.distinctterms as DistinctCount,
        c.distinctunknowns as UnknownCount,
        c.unknownpercent as UnknownPercent,
        c.status_distribution as StatusDistribution,
        case when completed_books.BkID is null then 0 else 1 end as IsCompleted

    FROM books b
    INNER JOIN languages ON LgID = b.BkLgID
    LEFT OUTER JOIN texts currtext ON currtext.TxID = BkCurrentTxID
    INNER JOIN (
        SELECT TxBkID, SUM(TxWordCount) as wc, COUNT(TxID) AS pagecount
        FROM texts
        GROUP BY TxBkID
    ) textcounts on textcounts.TxBkID = b.BkID
    LEFT OUTER JOIN bookstats c on c.BkID = b.BkID

    LEFT OUTER JOIN (
        SELECT BtBkID as BkID, GROUP_CONCAT(T2Text, ', ') AS taglist
        FROM
        (
            SELECT BtBkID, T2Text
            FROM booktags bt
            INNER JOIN tags2 t2 ON t2.T2ID = bt.BtT2ID
            ORDER BY T2Text
        ) tagssrc
        GROUP BY BtBkID
    ) AS tags ON tags.BkID = b.BkID

    left outer join (
      select texts.TxBkID as BkID
      from texts
      inner join (
        /* last page in each book */
        select TxBkID, max(TxOrder) as maxTxOrder from texts group by TxBkID
      ) last_page on last_page.TxBkID = texts.TxBkID and last_page.maxTxOrder = texts.TxOrder
      where TxReadDate is not null
    ) completed_books on completed_books.BkID = b.BkID

    WHERE b.BkArchived = {archived}
      and languages.LgParserType in ({ supported_parser_type_criteria() })
    """

    # Add "where" criteria for all the filters.
    language_id = parameters["filtLanguage"]
    if language_id == "null" or language_id == "undefined" or language_id is None:
        language_id = "0"
    language_id = int(language_id)
    if language_id != 0:
        base_sql += f" and LgID = {language_id}"

    connection = session.connection()
    return DataTablesSqliteQuery.get_data(base_sql, parameters, connection)



"""
Current settings tests.
"""

from lute.db import db
from lute.settings.current import (
    refresh_global_settings,
    current_settings,
    current_hotkeys,
)


def test_refresh_refreshes_current_settings(app_context):
    "Current settigns are loaded."
    if "backup_dir" in current_settings:
        del current_settings["backup_dir"]
    refresh_global_settings(db.session)
    assert "backup_dir" in current_settings, "loaded"


def test_hotkey_strings_mapped_to_name(app_context):
    "Hotkey key combo to name."
    refresh_global_settings(db.session)
    hotkey_names = current_hotkeys.values()
    assert "hotkey_Status5" in hotkey_names, "this is set by default"
    assert current_hotkeys["Digit5"] == "hotkey_Status5", "mapped"
    assert "" not in current_hotkeys, "No blank keyboard shortcuts"



"""
Reading rendering helpers.
"""

import itertools
import re
from sqlalchemy import text as sqltext

from lute.models.term import Term
from lute.parse.base import ParsedToken
from lute.read.render.calculate_textitems import get_textitems as calc_get_textitems
from lute.read.render.multiword_indexer import MultiwordTermIndexer

# from lute.utils.debug_helpers import DebugTimer


class Service:
    "Service."

    def __init__(self, session):
        self.session = session

    def find_all_Terms_in_string(self, s, language):  # pylint: disable=too-many-locals
        """
        Find all terms contained in the string s.

        For example
        - given s = "Here is a cat"
        - given terms in the db: [ "cat", "a cat", "dog" ]

        This would return the terms "cat" and "a cat".
        """
        cleaned = re.sub(r" +", " ", s)
        tokens = language.get_parsed_tokens(cleaned)
        return self._find_all_terms_in_tokens(tokens, language)

    def _get_multiword_terms(self, language):
        "Get all multiword terms."
        sql = sqltext(
            """
            SELECT WoID, WoTextLC FROM words
            WHERE WoLgID=:language_id and WoTokenCount>1
            """
        )
        sql = sql.bindparams(language_id=language.id)
        return self.session.execute(sql).all()

    def _find_all_multi_word_term_text_lcs_in_content(self, text_lcs, language):
        "Find multiword terms, return list of text_lcs."

        # There are a few ways of finding multi-word Terms
        # (with token_count > 1) in the content:
        #
        # 1. load each mword term text_lc via sql and check.
        # 2. using the model
        # 3. SQL with "LIKE"
        #
        # During reasonable test runs with my data, the times in seconds
        # for each are similar (~0.02, ~0.05, ~0.025).  This method is
        # only used for small amounts of data, and the user experience hit
        # is negligible, so I'll use the first method which IMO is the clearest
        # code.

        zws = "\u200B"  # zero-width space
        content = zws + zws.join(text_lcs) + zws

        # Method 1:
        reclist = self._get_multiword_terms(language)
        return [p[1] for p in reclist if f"{zws}{p[1]}{zws}" in content]

        ## # Method 2: use the model.
        ## contained_term_qry = self.session.query(Term).filter(
        ##     Term.language == language,
        ##     Term.token_count > 1,
        ##     func.instr(content, Term.text_lc) > 0,
        ## )
        ## return [r.text_lc for r in contained_term_qry.all()]

        ## # Method 3: Query with LIKE
        ## sql = sqltext(
        ##     """
        ##     SELECT WoTextLC FROM words
        ##     WHERE WoLgID=:lid and WoTokenCount>1
        ##     AND :content LIKE '%' || :zws || WoTextLC || :zws || '%'
        ##     """
        ## )
        ## sql = sql.bindparams(lid=language.id, content=content, zws=zws)
        ## recs = self.session.execute(sql).all()
        ## return [r[0] for r in recs]

    def _find_all_terms_in_tokens(self, tokens, language, kwtree=None):
        """
        Find all terms contained in the (ordered) parsed tokens tokens.

        For example
        - given tokens = "Here", " ", "is", " ", "a", " ", "cat"
        - given terms in the db: [ "cat", "a/ /cat", "dog" ]

        This would return the terms "cat" and "a/ /cat".

        Method:
        - build list of lowercase text in the tokens
        - append all multword term strings that exist in the content
        - query for Terms that exist in the list

        Note: this method only uses indexes for multiword terms, as any
        content analyzed is first parsed into tokens before being passed
        to this routine.  There's no need to search for single-word Terms
        in the tokenized strings, they can be found by a simple query.
        """

        # Performance: About half of the time in this routine is spent in
        # Step 1 (finding multiword terms), the rest in step 2 (the actual
        # query).
        # dt = DebugTimer("_find_all_terms_in_tokens", display=True)

        parser = language.parser
        text_lcs = [parser.get_lowercase(t.token) for t in tokens]

        # Step 1: get the multiwords in the content.
        if kwtree is None:
            mword_terms = self._find_all_multi_word_term_text_lcs_in_content(
                text_lcs, language
            )
        else:
            results = kwtree.search_all(text_lcs)
            mword_terms = [r[0] for r in results]
        # dt.step("filtered mword terms")

        # Step 2: load the Term objects.
        #
        # The Term fetch is actually performant -- there is no
        # real difference between loading the Term objects versus
        # loading raw data with SQL and getting dicts.
        #
        # Code for getting raw data:
        # param_keys = [f"w{i}" for i, _ in enumerate(text_lcs)]
        # keys_placeholders = ','.join([f":{k}" for k in param_keys])
        # param_dict = dict(zip(param_keys, text_lcs))
        # param_dict["langid"] = language.id
        # sql = sqltext(f"""SELECT WoID, WoTextLC FROM words
        #     WHERE WoLgID=:langid and WoTextLC in ({keys_placeholders})""")
        # sql = sql.bindparams(language.id, *text_lcs)
        # results = self.session.execute(sql, param_dict).fetchall()
        text_lcs.extend(mword_terms)
        tok_strings = list(set(text_lcs))
        terms_matching_tokens_qry = self.session.query(Term).filter(
            Term.text_lc.in_(tok_strings), Term.language == language
        )
        all_terms = terms_matching_tokens_qry.all()
        # dt.step("exec query")

        return all_terms

    def get_textitems(self, s, language, multiword_term_indexer=None):
        """
        Get array of TextItems for the string s.

        The multiword_term_indexer is a big performance boost, but takes
        time to initialize.
        """
        # Hacky reset of state of ParsedToken state.
        # _Shouldn't_ be needed but doesn't hurt, even if it's lame.
        ParsedToken.reset_counters()

        cleaned = re.sub(r" +", " ", s)
        tokens = language.get_parsed_tokens(cleaned)
        terms = self._find_all_terms_in_tokens(tokens, language, multiword_term_indexer)
        textitems = calc_get_textitems(tokens, terms, language, multiword_term_indexer)
        return textitems

    def get_multiword_indexer(self, language):
        "Return indexer loaded with all multiword terms."
        mw = MultiwordTermIndexer()
        for r in self._get_multiword_terms(language):
            mw.add(r[1])
        return mw

    def get_paragraphs(self, s, language):
        """
        Get array of arrays of TextItems for the given string s.

        This doesn't use an indexer, as it should only be used
        for a single page of text!
        """
        textitems = self.get_textitems(s, language)

        def _split_textitems_by_paragraph(textitems):
            "Split by ¶"
            ret = []
            curr_para = []
            for t in textitems:
                if t.text == "¶":
                    ret.append(curr_para)
                    curr_para = []
                else:
                    curr_para.append(t)
            if len(curr_para) > 0:
                ret.append(curr_para)
            return ret

        def _split_by_sentence_number(p):
            sentences = [
                list(sentence)
                for _, sentence in itertools.groupby(p, key=lambda t: t.sentence_number)
            ]
            for s in sentences:
                s[0].add_html_class("sentencestart")
            return sentences

        paras = [
            _split_by_sentence_number(list(sentences))
            for sentences in _split_textitems_by_paragraph(textitems)
        ]

        return paras



"""
Settings test.
"""

from unittest.mock import patch
import pytest
from lute.db import db
from lute.models.repositories import (
    UserSettingRepository,
    SystemSettingRepository,
    MissingUserSettingKeyException,
)
from tests.dbasserts import assert_sql_result


@pytest.fixture(name="us_repo")
def fixture_usersetting_repo(app_context):
    "Repo"
    r = UserSettingRepository(db.session)
    return r


@pytest.fixture(name="ss_repo")
def fixture_systemsetting_repo(app_context):
    "Repo"
    r = SystemSettingRepository(db.session)
    return r


def test_user_and_system_settings_do_not_intersect(us_repo, ss_repo):
    "A UserSetting is not available as a system setting."
    us_repo.set_value("backup_count", 42)
    db.session.commit()
    sql = "select StValue, StKeyType from settings where StKey = 'backup_count'"
    assert_sql_result(sql, ["42; user"], "loaded")
    u = us_repo.get_value("backup_count")
    assert u == "42", "found user setting"
    assert ss_repo.get_value("backup_count") is None, "not in system settings"


def test_save_and_retrieve_user_setting(us_repo):
    "Smoke tests."
    us_repo.set_value("backup_count", 42)
    sql = "select StValue from settings where StKey = 'backup_count'"
    assert_sql_result(sql, ["5"], "still default")

    db.session.commit()
    assert_sql_result(sql, ["42"], "now set")

    v = us_repo.get_value("backup_count")
    assert v == "42", "is string"


def test_missing_value_value_is_null(ss_repo):
    "Missing key = None."
    assert ss_repo.get_value("missing") is None, "missing key"


def test_smoke_last_backup(us_repo):
    "Check syntax only."
    v = us_repo.get_last_backup_datetime()
    assert v is None, "not set"

    us_repo.set_last_backup_datetime(42)
    v = us_repo.get_last_backup_datetime()
    assert v == 42, "set _and_ saved"


def test_get_backup_settings(us_repo):
    "Smoke test."
    us_repo.set_value("backup_dir", "blah")
    us_repo.set_value("backup_count", 12)
    us_repo.set_value("backup_warn", 0)
    db.session.commit()
    b = us_repo.get_backup_settings()
    assert b.backup_dir == "blah"
    assert b.backup_auto is True  # initial defaults
    assert b.backup_warn is False  # set to 0 above
    assert b.backup_count == 12
    assert b.last_backup_datetime is None


def test_time_since_last_backup_future(us_repo):
    """
    Check formatting when last backup is reported to be in the future.

    current time = 600, backup time = 900
    """
    b = us_repo.get_backup_settings()
    with patch("time.time", return_value=600):
        b.last_backup_datetime = 900
        assert b.time_since_last_backup is None


def test_time_since_last_backup_none(us_repo):
    """
    Check formatting when last backup is reported to be None.

    current time = 600, backup time = None
    """
    b = us_repo.get_backup_settings()
    with patch("time.time", return_value=600):
        b.last_backup_datetime = None
        assert b.time_since_last_backup is None


def test_time_since_last_backup_right_now(us_repo):
    """
    Check formatting when last backup is reported to be the same as current time.

    current time = 600, backup time = 600
    """
    b = us_repo.get_backup_settings()
    with patch("time.time", return_value=600):
        b.last_backup_datetime = 600
        assert b.time_since_last_backup == "0 seconds ago"


def test_time_since_last_backup_in_past(us_repo):
    """
    Check formatting when last backup is reported to be in the past.

    current time = 62899200, backup time = various
    """
    b = us_repo.get_backup_settings()
    now = 62899200
    with patch("time.time", return_value=now):
        b.last_backup_datetime = now - 45
        assert b.time_since_last_backup == "45 seconds ago"
        b.last_backup_datetime = now - 75
        assert b.time_since_last_backup == "1 minute ago"
        b.last_backup_datetime = now - 135
        assert b.time_since_last_backup == "2 minutes ago"
        b.last_backup_datetime = now - 3601
        assert b.time_since_last_backup == "1 hour ago"
        b.last_backup_datetime = now - 7201
        assert b.time_since_last_backup == "2 hours ago"
        b.last_backup_datetime = now - 86401
        assert b.time_since_last_backup == "1 day ago"
        b.last_backup_datetime = now - 172801
        assert b.time_since_last_backup == "2 days ago"
        b.last_backup_datetime = now - 604801
        assert b.time_since_last_backup == "1 week ago"
        b.last_backup_datetime = now - 15724801
        assert b.time_since_last_backup == "26 weeks ago"
        b.last_backup_datetime = now - 45360001
        assert b.time_since_last_backup == "75 weeks ago"


def test_get_or_set_user_setting_unknown_key_throws(us_repo):
    "Safety, ensure no typo for user settings."
    with pytest.raises(MissingUserSettingKeyException):
        us_repo.get_value("bad_key")
    with pytest.raises(MissingUserSettingKeyException):
        us_repo.set_value("bad_key", 17)



"""
Term service tests.
"""

import pytest
from lute.models.repositories import TermRepository
from lute.models.term import TermTag
from lute.db import db
from lute.term.service import Service, TermServiceException, BulkTermUpdateData
from tests.utils import add_terms
from tests.dbasserts import assert_sql_result

# Bulk parent update

# pylint: disable=unbalanced-tuple-unpacking


def assert_updated(termid, expected, msg=""):
    "Return a term as a string for fast asserts."
    t = TermRepository(db.session).find(termid)
    bad_keys = [
        k for k in expected.keys() if k not in ["parents", "status", "tags", "text"]
    ]
    assert len(bad_keys) == 0, "no bad keys " + ", ".join(bad_keys)
    if "text" in expected:
        assert expected["text"] == t.text, msg
    if "parents" in expected:
        assert expected["parents"] == sorted([p.text for p in t.parents]), msg
    if "status" in expected:
        assert expected["status"] == t.status, msg
    if "tags" in expected:
        assert expected["tags"] == sorted([tag.text for tag in t.term_tags]), msg


def _apply_updates(bud):
    "Apply BulkTermUpdateData bud."
    svc = Service(db.session)
    svc.apply_bulk_updates(bud)


def test_downcasing(app_context, spanish):
    "Add parent."
    [t, p] = add_terms(spanish, ["T", "P"])
    bud = BulkTermUpdateData(term_ids=[t.id], lowercase_terms=False, parent_id=p.id)
    _apply_updates(bud)
    expected = {"text": "T", "parents": ["P"], "status": 1, "tags": []}
    assert_updated(t.id, expected, "no downcasing")

    bud.lowercase_terms = True
    _apply_updates(bud)
    expected = {"text": "t", "parents": ["P"], "status": 1, "tags": []}
    assert_updated(t.id, expected, "t downcased")


def test_bulk_updates_all_terms_must_be_same_lang(app_context, spanish, english):
    "Update parent of term."
    t, p = add_terms(spanish, ["t", "p"])
    [e] = add_terms(english, ["e"])
    bud = BulkTermUpdateData(term_ids=[t.id, e.id], parent_id=p.id)
    svc = Service(db.session)
    with pytest.raises(TermServiceException, match="Terms not all the same language"):
        svc.apply_bulk_updates(bud)


def test_add_parent_by_id(app_context, spanish):
    "Add parent."
    [t, p] = add_terms(spanish, ["T", "p"])
    bud = BulkTermUpdateData(term_ids=[t.id], parent_id=p.id)
    _apply_updates(bud)
    expected = {"text": "T", "parents": ["p"], "status": 1, "tags": []}
    assert_updated(t.id, expected, "parent added")


def test_remove_all_parents(app_context, spanish):
    "Checkbox to remove all parents."
    [t, p] = add_terms(spanish, ["t", "p"])
    bud = BulkTermUpdateData(term_ids=[t.id], parent_id=p.id)
    _apply_updates(bud)
    expected = {"parents": ["p"], "status": 1, "tags": []}
    assert_updated(t.id, expected, "parent added")

    bud = BulkTermUpdateData(term_ids=[t.id], remove_parents=True)
    _apply_updates(bud)
    expected = {"parents": [], "status": 1, "tags": []}
    assert_updated(t.id, expected, "parent removed")


def test_add_parent_by_text_existing_parent(app_context, spanish):
    "Finds the existing parent by id."
    [t, _] = add_terms(spanish, ["t", "p"])
    bud = BulkTermUpdateData(term_ids=[t.id], parent_text="p")
    _apply_updates(bud)
    expected = {"parents": ["p"], "status": 1, "tags": []}
    assert_updated(t.id, expected, "parent added")


def test_add_parent_by_text_new_parent(app_context, spanish):
    "User can create a new parent in the tagify parent field."
    [t] = add_terms(spanish, ["t"])
    bud = BulkTermUpdateData(term_ids=[t.id], parent_text="newparent")
    _apply_updates(bud)
    expected = {"parents": ["newparent"], "status": 1, "tags": []}
    assert_updated(t.id, expected, "parent added")


def test_add_parent_text_ignored_if_id_present(app_context, spanish):
    "Just in case parent_id and parent_text are both sent."
    [t, p] = add_terms(spanish, ["t", "p"])
    bud = BulkTermUpdateData(term_ids=[t.id], parent_id=p.id, parent_text="ignored")
    _apply_updates(bud)
    expected = {"parents": ["p"], "status": 1, "tags": []}
    assert_updated(t.id, expected, "parent added")


def test_set_status(app_context, spanish):
    "Sanity check."
    [t] = add_terms(spanish, ["t"])
    bud = BulkTermUpdateData(term_ids=[t.id], change_status=True, status_value=4)
    _apply_updates(bud)
    expected = {"parents": [], "status": 4, "tags": []}
    assert_updated(t.id, expected, "status")


def test_set_status_skipped_if_change_status_false(app_context, spanish):
    "Sanity check."
    [t] = add_terms(spanish, ["t"])
    bud = BulkTermUpdateData(term_ids=[t.id], change_status=False, status_value=4)
    _apply_updates(bud)
    expected = {"parents": [], "status": 1, "tags": []}
    assert_updated(t.id, expected, "status unchanged")


def test_bulk_update_tag_add_and_remove_smoke_test(app_context, spanish):
    "Update parent of term."
    [t] = add_terms(spanish, ["t"])
    t.add_term_tag(TermTag("hello"))
    t.add_term_tag(TermTag("there"))
    db.session.add(t)
    db.session.commit()

    expected = {"parents": [], "status": 1, "tags": ["hello", "there"]}
    assert_updated(t.id, expected, "initial tags")

    bud = BulkTermUpdateData(
        term_ids=[t.id], add_tags=["hello", "cat"], remove_tags=["there", "dog"]
    )
    _apply_updates(bud)

    expected = {"parents": [], "status": 1, "tags": ["cat", "hello"]}
    assert_updated(t.id, expected, "tags added,removed")

    tagssql = "select TgText from tags order by TgText"
    expected_tags = ["cat", "hello", "there"]
    assert_sql_result(tagssql, expected_tags, "tag created and added if needed")



"""
Book mapping checks.
"""

from datetime import datetime
import pytest
from lute.models.book import Book, BookTag, TextBookmark, BookStats
from lute.read.service import Service
from lute.book.stats import Service as BookStatsService
from lute.db import db
from tests.dbasserts import assert_sql_result, assert_record_count_equals


@pytest.fixture(name="simple_book")
def fixture_simple_book(english):
    "Single page book with some associated objects."
    b = Book.create_book("hi", english, "SOME TEXT")
    b.texts[0].read_date = datetime.now()
    bt = BookTag.make_book_tag("hola")
    b.book_tags.append(bt)
    return b


def test_save_book(empty_db, simple_book):
    """
    Check book mappings.
    """
    b = simple_book
    db.session.add(b)
    db.session.commit()

    sql = "select BkID, BkTitle, BkLgID from books"
    assert_sql_result(sql, ["1; hi; 1"], "book")

    sql = "select TxID, TxBkID, TxText from texts"
    assert_sql_result(sql, ["1; 1; SOME TEXT"], "texts")

    sql = "select * from sentences"
    assert_sql_result(sql, ["1; 1; 1; /SOME/ /TEXT/; *"], "sentences")

    sql = "select * from booktags"
    assert_sql_result(sql, ["1; 1"], "booktags")

    sql = "select * from tags2"
    assert_sql_result(sql, ["1; hola; "], "tags2")


def test_delete_book(empty_db, simple_book):
    """
    Most tables should clear out after delete.
    """
    b = simple_book
    db.session.add(b)
    db.session.commit()

    service = Service(db.session)
    service.mark_page_read(b.id, 1, False)
    service.mark_page_read(b.id, 1, True)

    bss = BookStatsService(db.session)
    bss.refresh_stats()

    check_tables = ["books", "bookstats", "texts", "sentences", "booktags"]
    for t in check_tables:
        assert_record_count_equals(t, 1, f"{t} created")

    db.session.delete(b)
    db.session.commit()

    for t in check_tables:
        assert_record_count_equals(t, 0, f"{t} deleted")

    sql = "select * from tags2"
    assert_sql_result(sql, ["1; hola; "], "tags2 remain")

    sql = "select WrTxID, WrWordCount from wordsread"
    assert_sql_result(sql, ["None; 2", "None; 2"], "words read remains.")


def test_save_and_delete_created_book(english):
    """
    Verify book orm mappings.
    """
    content = "Some text here. Some more text"
    b = Book.create_book("test", english, content, 3)
    db.session.add(b)
    db.session.commit()
    sql = f"select TxOrder, TxText from texts where TxBkID = {b.id}"
    expected = ["1; Some text here.", "2; Some more text"]
    assert_sql_result(sql, expected, "texts")

    db.session.delete(b)
    db.session.commit()
    assert_sql_result(sql, [], "texts deleted")


def test_load_book_loads_lang(empty_db, simple_book):
    """
    Check book mappings.
    """
    b = simple_book
    db.session.add(b)
    db.session.commit()

    findbook = db.session.query(Book).filter(Book.title == "hi").first()
    assert findbook.title == "hi", "title"
    assert findbook.language.name == "English", "check lang"

    for b in db.session.query(Book).all():
        assert b.language is not None, "have lang object"

    books_to_update = (
        db.session.query(Book)
        .filter(~Book.id.in_(db.session.query(BookStats.BkID)))
        .all()
    )
    for b in books_to_update:
        assert b.language is not None, "have lang object"


def test_delete_book_cascade_deletes_bookmarks(empty_db, simple_book):
    """
    All associated TextBookmark(s) should be deleted when their
    associated Book entry is deleted.
    """
    b = simple_book

    TextBookmark(title="hello", text=b.texts[0])
    db.session.add(b)
    db.session.commit()

    sql = "select BkID, BkTitle, BkLgID from books"
    assert_sql_result(sql, ["1; hi; 1"], "book")
    sql = "select TxID, TxBkID, TxText from texts"
    assert_sql_result(sql, ["1; 1; SOME TEXT"], "texts")
    sql = "select TbID, TbTxID, TbTitle from textbookmarks"
    assert_sql_result(sql, ["1; 1; hello"], "bookmarks")

    db.session.delete(b)
    db.session.commit()

    for t in ["books", "texts", "textbookmarks"]:
        assert_record_count_equals(t, 0, f"{t} deleted")



"""
/language endpoints.
"""

from sqlalchemy import text
from sqlalchemy.exc import IntegrityError
from flask import Blueprint, current_app, render_template, redirect, url_for, flash
from lute.models.language import Language
from lute.models.repositories import UserSettingRepository
from lute.language.service import Service
from lute.language.forms import LanguageForm
from lute.db import db
from lute.parse.registry import supported_parsers

bp = Blueprint("language", __name__, url_prefix="/language")


@bp.route("/index")
def index():
    """
    List all languages, with book and term counts.
    """

    # Using plain sql, easier to get bulk quantities.
    sql = """
    select LgID, LgName, book_count, term_count from languages
    left outer join (
      select BkLgID, count(BkLgID) as book_count from books
      group by BkLgID
    ) bc on bc.BkLgID = LgID
    left outer join (
      select WoLgID, count(WoLgID) as term_count from words
      where WoStatus != 0
      group by WoLgID
    ) tc on tc.WoLgID = LgID
    order by LgName
    """
    result = db.session.execute(text(sql)).all()
    languages = [
        {"LgID": row[0], "LgName": row[1], "book_count": row[2], "term_count": row[3]}
        for row in result
    ]
    return render_template("language/index.html", language_data=languages)


def _handle_form(language, form) -> bool:
    """
    Handle the language form processing.

    Returns True if validated and saved.
    """
    result = False

    if form.validate_on_submit():
        try:
            form.populate_obj(language)
            current_app.db.session.add(language)
            current_app.db.session.commit()
            flash(f"Language {language.name} updated", "success")
            result = True
        except IntegrityError as e:
            current_app.db.session.rollback()
            msg = e.orig
            if "languages.LgName" in f"{e.orig}":
                msg = f"Language {form.name.data} already exists."
            flash(msg, "error")

    return result


def _add_hidden_dictionary_template_entry(form):
    "Add a dummy placeholder dictionary to be used as a template."
    # Add a dummy dictionary entry with dicturi __TEMPLATE__.
    #
    # This entry is used as a "template" when adding a new dictionary
    # to the list of dictionaries (see templates/language/_form.html).
    # This is the easiest way to ensure that new dictionary entries
    # have the correct controls.
    #
    # This dummy entry is not rendered on the form, or submitted
    # when the form is submitted.  Search for __TEMPLATE__ in
    # templates/language/_form.html to see where it is handled.
    form.dictionaries.append_entry({"dicturi": "__TEMPLATE__"})


def _dropdown_parser_choices():
    "Get dropdown list of parser type name to name."
    return [(a[0], a[1].name()) for a in supported_parsers()]


@bp.route("/edit/<int:langid>", methods=["GET", "POST"])
def edit(langid):
    """
    Edit a language.
    """
    language = db.session.get(Language, langid)

    if not language:
        flash(f"Language {langid} not found", "danger")
        return redirect(url_for("language.index"))

    form = LanguageForm(obj=language)
    form.parser_type.choices = _dropdown_parser_choices()

    if _handle_form(language, form):
        return redirect("/")

    _add_hidden_dictionary_template_entry(form)

    return render_template("language/edit.html", form=form, language=language)


@bp.route("/new", defaults={"langname": None}, methods=["GET", "POST"])
@bp.route("/new/<string:langname>", methods=["GET", "POST"])
def new(langname):
    """
    Create a new language.
    """
    service = Service(db.session)
    predefined = service.supported_predefined_languages()
    language = Language()
    if langname is not None:
        candidates = [lang for lang in predefined if lang.name == langname]
        if len(candidates) == 1:
            language = candidates[0]

    form = LanguageForm(obj=language)
    form.parser_type.choices = _dropdown_parser_choices()

    if _handle_form(language, form):
        # New language, so show everything b/c user should re-choose
        # the default.
        #
        # Reason for this: a user may start off with just language X,
        # so the current_language_id is set to X.id.  If the user then
        # adds language Y, the filter stays on X, which may be
        # disconcerting/confusing.  Forcing a reselect is painless and
        # unambiguous.
        repo = UserSettingRepository(db.session)
        repo.set_value("current_language_id", 0)
        db.session.commit()
        return redirect("/")

    _add_hidden_dictionary_template_entry(form)

    return render_template(
        "language/new.html", form=form, language=language, predefined=predefined
    )


@bp.route("/delete/<int:langid>", methods=["POST"])
def delete(langid):
    """
    Delete a language.
    """
    language = db.session.get(Language, langid)
    if not language:
        flash(f"Language {langid} not found")
    db.session.delete(language)
    db.session.commit()
    return redirect(url_for("language.index"))


@bp.route("/list_predefined", methods=["GET"])
def list_predefined():
    "Show supported predefined languages that are not already in the db."
    service = Service(db.session)
    predefined = service.supported_predefined_languages()
    existing_langs = db.session.query(Language).all()
    existing_names = [l.name for l in existing_langs]
    new_langs = [p for p in predefined if p.name not in existing_names]
    return render_template("language/list_predefined.html", predefined=new_langs)


@bp.route("/load_predefined/<langname>", methods=["GET"])
def load_predefined(langname):
    "Load a predefined language and its stories."
    service = Service(db.session)
    lang_id = service.load_language_def(langname)
    repo = UserSettingRepository(db.session)
    repo.set_value("current_language_id", lang_id)
    db.session.commit()
    flash(f"Loaded {langname} and sample book(s)")
    return redirect("/")



"""
Simple CLI commands.
"""

import click
from flask import Blueprint

from lute.cli.language_term_export import generate_language_file, generate_book_file
from lute.cli.import_books import import_books_from_csv

bp = Blueprint("cli", __name__)


@bp.cli.command("hello")
def hello():
    "Say hello -- proof-of-concept CLI command only."
    msg = """
    Hello there!

    This is the Lute cli.

    There may be some experimental scripts here ...
    nothing that will change or damage your Lute data,
    but the CLI may change.

    Thanks for looking.
    """
    print(msg)


@bp.cli.command("language_export")
@click.argument("language")
@click.argument("output_path")
def language_export(language, output_path):
    """
    Get all terms from all books in the language, and write a
    data file of term frequencies and children.
    """
    generate_language_file(language, output_path)


@bp.cli.command("book_term_export")
@click.argument("bookid")
@click.argument("output_path")
def book_term_export(bookid, output_path):
    """
    Get all terms for the given book, and write a
    data file of term frequencies and children.
    """
    generate_book_file(bookid, output_path)


@bp.cli.command("import_books_from_csv")
@click.option(
    "--commit",
    is_flag=True,
    help="""
    Commit the changes to the database. If not set, import in dry-run mode. A
    list of changes will be printed out but not applied.
""",
)
@click.option(
    "--tags",
    default="",
    help="""
    A comma-separated list of tags to apply to all books.
""",
)
@click.option(
    "--language",
    default="",
    help="""
    The name of the default language to apply to each book, as it appears in
    your language settings. If unset, the language must be indicated in the
    "language" column of the CSV file.
""",
)
@click.argument("file")
def import_books_from_csv_cmd(language, file, tags, commit):
    """
    Import books from a CSV file.

    The CSV file must have a header row with the following, case-sensitive,
    column names. The order of the columns does not matter. The CSV file may
    include additional columns, which will be ignored.

      - title: the title of the book

      - text: the text of the book

      - language: [optional] the name of the language of book, as it appears in
      your language settings. If unspecified, the language specified on the
      command line (using the --language option) will be used.

      - url: [optional] the source URL for the book

      - tags: [optional] a comma-separated list of tags to apply to the book
      (e.g., "audiobook,beginner")

      - audio: [optional] the path to the audio file of the book. This should
      either be an absolute path, or a path relative to the CSV file.

      - bookmarks: [optional] a semicolon-separated list of audio bookmark
      positions, in seconds (decimals permitted; e.g., "12.34;42.89;89.00").
    """
    tags = list(tags.split(",")) if tags else []
    import_books_from_csv(file, language, tags, commit)



"""
Functions to manage demo database.

Lute db comes pre-loaded with some demo data.  User can view Tutorial,
wipe data, etc.

The db settings table contains a record, StKey = 'IsDemoData', if the
data is demo.
"""

from sqlalchemy import text
from lute.language.service import Service as LanguageService
from lute.book.model import Repository
from lute.book.stats import Service as StatsService
from lute.models.repositories import SystemSettingRepository, LanguageRepository
import lute.db.management


class Service:
    "Demo database service."

    def __init__(self, session):
        self.session = session

    def _demo_languages(self):
        """
        Demo languages to be loaded for new users.
        Also loaded during tests.
        """
        return [
            "Arabic",
            "Classical Chinese",
            "Czech",
            "English",
            "French",
            "German",
            "Greek",
            "Hindi",
            "Japanese",
            "Russian",
            "Sanskrit",
            "Spanish",
            "Turkish",
        ]

    def set_load_demo_flag(self):
        "Set the flag."
        repo = SystemSettingRepository(self.session)
        repo.set_value("LoadDemoData", True)
        self.session.commit()

    def remove_load_demo_flag(self):
        "Set the flag."
        repo = SystemSettingRepository(self.session)
        repo.delete_key("LoadDemoData")
        self.session.commit()

    def _flag_exists(self, flagname):
        "True if flag exists, else false."
        repo = SystemSettingRepository(self.session)
        return repo.key_exists(flagname)

    def should_load_demo_data(self):
        return self._flag_exists("LoadDemoData")

    def contains_demo_data(self):
        return self._flag_exists("IsDemoData")

    def remove_flag(self):
        """
        Remove IsDemoData setting.
        """
        if not self.contains_demo_data():
            raise RuntimeError("Can't delete non-demo data.")

        repo = SystemSettingRepository(self.session)
        repo.delete_key("IsDemoData")
        self.session.commit()

    def tutorial_book_id(self):
        """
        Return the book id of the tutorial.
        """
        if not self.contains_demo_data():
            return None
        sql = """select BkID from books
        inner join languages on LgID = BkLgID
        where LgName = 'English' and BkTitle = 'Tutorial'
        """
        r = self.session.execute(text(sql)).first()
        if r is None:
            return None
        return int(r[0])

    def delete_demo_data(self):
        """
        If this is a demo, wipe everything.
        """
        if not self.contains_demo_data():
            raise RuntimeError("Can't delete non-demo data.")
        self.remove_flag()
        lute.db.management.delete_all_data(self.session)

    # Loading demo data.

    def load_demo_languages(self):
        """
        Load selected predefined languages, if they're supported.

        This method will also be called during acceptance tests, so it's public.
        """
        demo_langs = self._demo_languages()
        service = LanguageService(self.session)
        langs = [service.get_language_def(langname).language for langname in demo_langs]
        supported = [lang for lang in langs if lang.is_supported]
        for lang in supported:
            self.session.add(lang)
        self.session.commit()

    def load_demo_stories(self):
        "Load the stories for any languages already loaded."
        demo_langs = self._demo_languages()
        service = LanguageService(self.session)
        langdefs = [service.get_language_def(langname) for langname in demo_langs]

        langrepo = LanguageRepository(self.session)
        langdefs = [
            d
            for d in langdefs
            if d.language.is_supported
            and langrepo.find_by_name(d.language.name) is not None
        ]

        r = Repository(self.session)
        for d in langdefs:
            for b in d.books:
                r.add(b)
        r.commit()

        repo = SystemSettingRepository(self.session)
        repo.set_value("IsDemoData", True)
        self.session.commit()

        svc = StatsService(self.session)
        svc.refresh_stats()

    def _db_has_data(self):
        "True of the db contains any language data."
        sql = "select LgID from languages limit 1"
        r = self.session.execute(text(sql)).first()
        return r is not None

    def load_demo_data(self):
        """
        Load the data.
        """
        if self._db_has_data():
            self.remove_load_demo_flag()
            return

        repo = SystemSettingRepository(self.session)
        do_load = repo.get_value("LoadDemoData")
        if do_load is None:
            # Only load if flag is explicitly set.
            return

        do_load = bool(int(do_load))
        if not do_load:
            return

        self.load_demo_languages()
        self.load_demo_stories()
        self.remove_load_demo_flag()
        repo.set_value("IsDemoData", True)
        self.session.commit()
